# Representing a scalar

# You can construct and manipulate a population of neurons (ensemble) in nengo.
# This model shows how the activity of neural populations can be thought
# of as representing a mathematical variable (a scalar value).

# The model has paramenters as described in the book and uses a single
# population (ensemble) of 100 LIF neurons. Note that the default max rates
# in Nengo 2.0 are (200, 400), so you have to explicitly specify them to be
# (100, 200) to create the model with the same parameters as described in the
# book. Moreover the 'Node Factory' feature of ensembles mentioned in the book
# maps to the 'neuron_type' in Nengo 2.0 which is set to LIF by default. The
# default values of tauRC, tauRef, radius and intercepts in Nengo 2.0 are the
# same as those mentioned in the book.

# Press the play button to run the simulation.
# The graph on the top left shows the input and the graph on the top right
# shows the the decoded value of the neural spiking (a linearly decoded estimate
# of the input). The graph on the bottom right shows the spike raster which is
# the spiking output of the neuron population (x).

import nengo

# Setup the environment
import numpy as np
from nengo.dists import Uniform

# Create the network object to which we can add ensembles, connections, etc.
model = nengo.Network(label="Many Neurons")

with model:
    # Input sine wave with range 1, freq of 16 rad/s
    stim = nengo.Node(lambda t: np.sin(16 * t), label="input")

    # Input sine wave with range increased to 4
    # stim = nengo.Node(lambda t: 4 * np.sin(16 * t), label="input")

    # Ensemble with 100 LIF neurons
    x = nengo.Ensemble(100, dimensions=1, max_rates=Uniform(100, 200))

    # Connecting input to ensemble
    nengo.Connection(stim, x)


# Increasing the range of Input

# You have seen that the population of neurons does a reasonably good job of
# representing the input. However, neurons cannot represent arbitrary values
# well and you can verify this by increasing the range of the input to 4
# ( input = nengo.Node(lambda t: 4 * np.sin(16 t)) ). You will observe the same
# saturation effects as described in the book, showing that the neurons do a
# much better job at representing information within the defined radius.



#!/usr/bin/env python
#
# This code is part of Grandalf
# Copyright (C) 2008 Axel Tillequin (bdcht3@gmail.com) and others
# published under GPLv2 license or EPLv1 license
# Contributor(s): Axel Tillequin, Fabio Zadrozny

import math
from math import atan2, cos, sin, sqrt

import numpy
from numpy import array

try:
    import ply.lex as lex
    import ply.yacc as yacc

    _has_ply = True
except ImportError:
    _has_ply = False

# ------------------------------------------------------------------------------
class Poset(object):
    def __init__(self, L):
        self.o = []
        s = set()
        for obj in L:
            if not obj in s:
                self.o.append(obj)
                s.add(obj)
            else:
                print(
                    "warning: obj was already added in poset at index %d" % s.index(obj)
                )
        self.s = s

    def __repr__(self):
        return "poset(%r)" % (self.o,)

    def __str__(self):
        f = "%%%dd" % len(str(len(self.o)))
        s = []
        for i, x in enumerate(self.o):
            s.append(f % i + ".| %s" % repr(x))
        return "\n".join(s)

    def add(self, obj):
        if obj not in self.s:
            self.o.append(obj)
            self.s.add(obj)

    def remove(self, obj):
        if obj in self.s:
            self.o.remove(obj)
            self.s.remove(obj)

    def index(self, obj):
        return self.o.index(obj)

    def __len__(self):
        return len(self.o)

    def __iter__(self):
        for obj in self.o:
            yield obj

    def __cmp__(self, other):
        return cmp(other.s, self.s)

    def __eq__(self, other):
        return other.s == self.s

    def __ne__(self, other):
        return other.s != self.s

    def copy(self):
        return Poset(self.o)

    __copy__ = copy

    def deepcopy(self):
        from copy import deepcopy

        L = deepcopy(self.o)
        return Poset(L)

    def __or__(self, other):
        return self.union(other)

    def union(self, other):
        return Poset(self.o + other.o)

    def update(self, other):
        for obj in other:
            if obj not in self:
                self.add(obj)

    def __and__(self, other):
        return self.intersection(other)

    def intersection(self, other):
        return Poset(self.s.intersection(other.s))

    def __xor__(self, other):
        return self.symmetric_difference(other)

    def symmetric_difference(self, other):
        return Poset(self.s ^ other.s)

    def __sub__(self, other):
        return self.difference(other)

    def difference(self, other):
        return Poset(self.s - other.s)

    def __contains__(self, obj):
        return obj in self.s

    def issubset(self, other):
        return self.s.issubset(other.s)

    def issuperset(self, other):
        return self.s.issuperset(other.s)

    __le__ = issubset
    __ge__ = issuperset

    def __lt__(self, other):
        return self <= other and len(self) != len(other)

    def __gt__(self, other):
        return self >= other and len(self) != len(other)


#  rand_ortho1 returns a numpy.array representing
#  a random normalized n-dimension vector orthogonal to (1,1,1,...,1).
def rand_ortho1(n):
    from random import SystemRandom

    r = SystemRandom()
    pos = [r.random() for x in xrange(n)]
    s = sum(pos)

    # Note: returning only positive values (if we had only negative values
    # an exception such as:
    # inf X:\grandalf\grandalf\layouts.py:784: RuntimeWarning: divide by zero encountered in double_scalars
    #   sfactor = 1.0/max(y.max(),x.max())
    # Could be found (randomly happening in test-layouts:test_splines).
    v = abs(array(pos, dtype=float) - (s / len(pos)))
    norm = sqrt(sum(v * v))
    return v / norm


# ------------------------------------------------------------------------------
# TODO:  this was imported here from masr, but since we have
#  here access to numpy.array, we could use it for vectors operations.
def intersect2lines(p1, p2, p3, p4):
    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3
    x4, y4 = p4
    b = (x2 - x1, y2 - y1)
    d = (x4 - x3, y4 - y3)
    det = b[0] * d[1] - b[1] * d[0]
    if det == 0:
        return None
    c = (x3 - x1, y3 - y1)
    t = float(c[0] * b[1] - c[1] * b[0]) / (det * 1.0)
    if t < 0.0 or t > 1.0:
        return None
    t = float(c[0] * d[1] - c[1] * d[0]) / (det * 1.0)
    if t < 0.0 or t > 1.0:
        return None
    x = x1 + t * b[0]
    y = y1 + t * b[1]
    return (x, y)


# ------------------------------------------------------------------------------
#  intersectR returns the intersection point between the Rectangle
#  (w,h) that characterize the view object and the line that goes
#  from the views' object center to the 'topt' point.
def intersectR(view, topt):
    # we compute intersection in local views' coord:
    # center of view is obviously :
    x1, y1 = 0, 0
    # endpoint in view's coord:
    x2, y2 = topt[0] - view.xy[0], topt[1] - view.xy[1]
    # bounding box:
    bbx2 = view.w / 2
    bbx1 = -bbx2
    bby2 = view.h / 2
    bby1 = -bby2
    # all 4 segments of the bb:
    S = [
        ((x1, y1), (x2, y2), (bbx1, bby1), (bbx2, bby1)),
        ((x1, y1), (x2, y2), (bbx2, bby1), (bbx2, bby2)),
        ((x1, y1), (x2, y2), (bbx1, bby2), (bbx2, bby2)),
        ((x1, y1), (x2, y2), (bbx1, bby2), (bbx1, bby1)),
    ]
    # check intersection with each seg:
    for segs in S:
        xy = intersect2lines(*segs)
        if xy != None:
            x, y = xy
            # return global coord:
            x += view.xy[0]
            y += view.xy[1]
            return (x, y)
    # there can't be no intersection unless the endpoint was
    # inside the bb !
    raise ValueError(
        "no intersection found (point inside ?!). view: %s topt: %s" % (view, topt)
    )


# ------------------------------------------------------------------------------
def getangle(p1, p2):
    x1, y1 = p1
    x2, y2 = p2
    theta = atan2(y2 - y1, x2 - x1)
    return theta


# ------------------------------------------------------------------------------
#  intersectC returns the intersection point between the Circle
#  of radius r and centered on views' position with the line
#  to the 'topt' point.
def intersectC(view, r, topt):
    theta = getangle(view.xy, topt)
    x = int(cos(theta) * r)
    y = int(sin(theta) * r)
    return (x, y)


def median_wh(views):
    mw = [v.w for v in views]
    mh = [v.h for v in views]
    mw.sort()
    mh.sort()
    return (mw[len(mw) // 2], mh[len(mh) // 2])


# ------------------------------------------------------------------------------
#  setcurve returns the spline curve that path through the list of points P.
#  The spline curve is a list of cubic bezier curves (nurbs) that have
#  matching tangents at their extreme points.
#  The method considered here is taken from "The NURBS book" (Les A. Piegl,
#  Wayne Tiller, Springer, 1997) and implements a local interpolation rather
#  than a global interpolation.
def setcurve(e, pts, tgs=None):
    P = map(array, pts)
    n = len(P)
    # tangent estimation
    if tgs:
        assert len(tgs) == n
        T = map(array, tgs)
        Q = [P[k + 1] - P[k] for k in range(0, n - 1)]
    else:
        Q, T = tangents(P, n)
    splines = []
    for k in xrange(n - 1):
        t = T[k] + T[k + 1]
        a = 16.0 - (t.dot(t))
        b = 12.0 * (Q[k].dot(t))
        c = -36.0 * Q[k].dot(Q[k])
        D = (b * b) - 4.0 * a * c
        assert D >= 0
        sd = sqrt(D)
        s1, s2 = (-b - sd) / (2.0 * a), (-b + sd) / (2.0 * a)
        s = s2
        if s1 >= 0:
            s = s1
        C0 = tuple(P[k])
        C1 = tuple(P[k] + (s / 3.0) * T[k])
        C2 = tuple(P[k + 1] - (s / 3.0) * T[k + 1])
        C3 = tuple(P[k + 1])
        splines.append([C0, C1, C2, C3])
    return splines


def tangents(P, n):
    assert n >= 2
    Q = []
    T = []
    for k in xrange(0, n - 1):
        q = P[k + 1] - P[k]
        t = q / sqrt(q.dot(q))
        Q.append(q)
        T.append(t)
    T.append(t)
    return (Q, T)


# ------------------------------------------------------------------------------
def setroundcorner(e, pts):
    P = map(array, pts)
    n = len(P)
    Q, T = tangents(P, n)
    c0 = P[0]
    t0 = T[0]
    k0 = 0
    splines = []
    k = 1
    while k < n:
        z = abs(t0[0] * T[k][1] - (t0[1] * T[k][0]))
        if z < 1.0e-6:
            k += 1
            continue
        if (k - 1) > k0:
            splines.append([c0, P[k - 1]])
        if (k + 1) < n:
            splines.extend(setcurve(e, [P[k - 1], P[k + 1]], tgs=[T[k - 1], T[k + 1]]))
        else:
            splines.extend(setcurve(e, [P[k - 1], P[k]], tgs=[T[k - 1], T[k]]))
            break
        if (k + 2) < n:
            c0 = P[k + 1]
            t0 = T[k + 1]
            k0 = k + 1
            k += 2
        else:
            break
    return splines or [[P[0], P[-1]]]


# ------------------------------------------------------------------------------
# LALR(1) parser for Graphviz dot file format.
class Dot:

    _reserved = (
        "strict",
        "graph",
        "digraph",
        "subgraph",
        "node",
        "edge",
    )
    _tokens = (
        "regulars",
        "string",
        "html",
        "comment",
    ) + _reserved

    _literals = [",", ";", "-", ">", "=", ":", "[", "]", "{", "}"]

    class Lexer(object):
        def __init__(self):
            self.whitespace = "\0\t\n\f\r "
            self.reserved = Dot._reserved
            self.tokens = Dot._tokens
            self.literals = Dot._literals
            self.t_ignore = self.whitespace

        def t_regulars(self, t):
            r"[-]?[\w.]+"
            v = t.value.lower()
            if v in self.reserved:
                t.type = v
                return t
            # check numeric string
            if v[0].isdigit() or v[0] in ["-", "."]:
                try:
                    float(v)
                except ValueError:
                    print("invalid numeral token: %s" % v)
                    raise SyntaxError
            elif "." in v:  # forbidden in non-numeric
                raise SyntaxError
            return t

        def t_comment_online(self, t):
            r"(//(.*)\n)|\\\n"
            pass

        def t_comment_macro(self, t):
            r"(\#(.*)\n)"
            pass

        def t_comment_multline(self, t):
            r"(/\*)"
            start = t.lexer.lexpos
            t.lexer.lexpos = t.lexer.lexdata.index("*/", start) + 2

        def t_string(self, t):
            r'"'
            start = t.lexer.lexpos - 1
            i = t.lexer.lexdata.index('"', start + 1)
            while t.lexer.lexdata[i - 1] == "\\":
                i = t.lexer.lexdata.index('"', i + 1)
            t.value = t.lexer.lexdata[start : i + 1]
            t.lexer.lexpos = i + 1
            return t

        def t_html(self, t):
            r"<"
            start = t.lexer.lexpos - 1
            level = 1
            i = start + 1
            while level > 0:
                c = t.lexer.lexdata[i]
                if c == "<":
                    level += 1
                if c == ">":
                    level -= 1
                i += 1
            t.value = t.lexer.lexdata[start:i]
            t.lexer.lexpos = i
            return t

        def t_ANY_error(self, t):
            print("Illegal character '%s'" % t.value[0])
            t.lexer.skip(1)

        def build(self, **kargs):
            if _has_ply:
                self._lexer = lex.lex(module=self, **kargs)

        def test(self, data):
            self._lexer.input(data)
            while 1:
                tok = self._lexer.token()
                if not tok:
                    break
                print(tok)

    # Classes for the AST returned by Parser:
    class graph(object):
        def __init__(self, name, data, strict=None, direct=None):
            self.name = name
            self.strict = strict
            self.direct = direct
            self.nodes = {}
            self.edges = []
            self.subgraphs = []
            self.attr = {}
            eattr = {}
            nattr = {}
            for x in data:  # data is a statements (list of stmt)
                # x is a stmt, ie one of:
                # a graph object (subgraph)
                # a attr object (graph/node/edge attributes)
                # a dict object (ID=ID)
                # a node object
                # a list of edges
                if isinstance(x, Dot.graph):
                    self.subgraphs.append(x)
                elif isinstance(x, Dot.attr):
                    if x.type == "graph":
                        self.attr.update(x.D)
                    elif x.type == "node":
                        nattr.update(x.D)
                    elif x.type == "edge":
                        eattr.update(x.D)
                    else:
                        raise TypeError("invalid attribute type")
                elif isinstance(x, dict):
                    self.attr.update(x)
                elif isinstance(x, Dot.node):
                    x.attr.update(nattr)
                    self.nodes[x.name] = x
                else:
                    for e in x:
                        e.attr.update(eattr)
                        self.edges.append(e)
                        for n in [e.n1, e.n2]:
                            if isinstance(n, Dot.graph):
                                continue
                            if n.name not in self.nodes:
                                n.attr.update(nattr)
                                self.nodes[n.name] = n

        def __repr__(self):
            u = "<%s instance at %x, name: %s, %d nodes>" % (
                self.__class__,
                id(self),
                self.name,
                len(self.nodes),
            )
            return u.encode("utf-8")

    class attr(object):
        def __init__(self, type, D):
            self.type = type
            self.D = D

    class edge(object):
        def __init__(self, n1, n2):
            self.n1 = n1
            self.n2 = n2
            self.attr = {}

    class node(object):
        def __init__(self, name, port=None):
            self.name = name
            self.port = port
            self.attr = {}

    class Parser(object):
        def __init__(self):
            self.tokens = Dot._tokens

        def __makelist(self, p):
            N = len(p)
            if N > 2:
                L = p[1]
                L.append(p[N - 1])
            else:
                L = []
                if N > 1:
                    L.append(p[N - 1])
            p[0] = L

        def p_Data(self, p):
            """Data : Data Graph
            | Graph"""
            self.__makelist(p)

        def p_Graph_strict(self, p):
            """Graph : strict graph name Block"""
            p[0] = Dot.graph(name=p[3], data=p[4], strict=1, direct=0)
            # print 'Dot.Parser: graph object %s created'%p[0].name

        def p_Graph_graph(self, p):
            """Graph : graph name Block"""
            p[0] = Dot.graph(name=p[2], data=p[3], strict=0, direct=0)

        def p_Graph_strict_digraph(self, p):
            """Graph : strict digraph name Block"""
            p[0] = Dot.graph(name=p[3], data=p[4], strict=1, direct=1)

        def p_Graph_digraph(self, p):
            """Graph : digraph name Block"""
            p[0] = Dot.graph(name=p[2], data=p[3], strict=0, direct=1)

        def p_ID(self, p):
            """ID : regulars
            | string
            | html"""
            p[0] = p[1]

        def p_name(self, p):
            """name : ID
            |"""
            if len(p) == 1:
                p[0] = ""
            else:
                p[0] = p[1]

        def p_Block(self, p):
            """Block : '{' statements '}'"""
            p[0] = p[2]

        def p_statements(self, p):
            """statements : statements stmt
            | stmt
            |"""
            self.__makelist(p)

        def p_stmt(self, p):
            """stmt : stmt ';'"""
            p[0] = p[1]

        def p_comment(self, p):
            """stmt : comment"""
            pass  # comment tokens are not outputed by lexer anyway

        def p_stmt_sub(self, p):
            """stmt : sub"""
            p[0] = p[1]

        def p_subgraph(self, p):
            """sub : subgraph name Block
            | Block"""
            N = len(p)
            if N > 2:
                ID = p[2]
            else:
                ID = ""
            p[0] = Dot.graph(name=ID, data=p[N - 1], strict=0, direct=0)

        def p_stmt_assign(self, p):
            """stmt : affect"""
            p[0] = p[1]

        def p_affect(self, p):
            """affect : ID '=' ID"""
            p[0] = dict([(p[1], p[3])])

        def p_stmt_lists(self, p):
            """stmt : graph attrs
            | node  attrs
            | edge  attrs"""
            p[0] = Dot.attr(p[1], p[2])

        def p_attrs(self, p):
            """attrs : attrs attrl
            | attrl"""
            if len(p) == 3:
                p[1].update(p[2])
            p[0] = p[1]

        def p_attrl(self, p):
            """attrl : '[' alist ']'"""
            L = {}
            for a in p[2]:
                if isinstance(a, dict):
                    L.update(a)
                else:
                    L[a] = "true"
            p[0] = L

        def p_alist_comma(self, p):
            """alist : alist ',' alist"""
            p[1].extend(p[3])
            p[0] = p[1]

        def p_alist_affect(self, p):
            """alist : alist affect
            | alist ID
            | affect
            | ID
            |"""
            self.__makelist(p)

        def p_stmt_E_attrs(self, p):
            """stmt : E attrs"""
            for e in p[1]:
                e.attr = p[2]
            p[0] = p[1]

        def p_stmt_N_attrs(self, p):
            """stmt : N attrs"""
            p[1].attr = p[2]
            p[0] = p[1]

        def p_stmt_EN(self, p):
            """stmt : E
            | N"""
            p[0] = p[1]

        def p_E(self, p):
            """E : E   link
            | elt link"""
            try:
                L = p[1]
                L.append(Dot.edge(L[-1].n2, p[2]))
            except:
                L = []
                L.append(Dot.edge(p[1], p[2]))
            p[0] = L

        def p_elt(self, p):
            """elt : N
            | sub"""
            p[0] = p[1]

        def p_link(self, p):
            """link : '-' '>' elt
            | '-' '-' elt"""
            p[0] = p[3]

        def p_N_port(self, p):
            """N : ID port"""
            p[0] = Dot.node(p[1], port=p[2])

        def p_N(self, p):
            """N : ID"""
            p[0] = Dot.node(p[1])

        def p_port(self, p):
            """port : ':' ID"""
            p[0] = p[2]

        def p_port2(self, p):
            """port : port port"""
            assert p[2] in ["n", "ne", "e", "se", "s", "sw", "w", "nw", "c", "_"]
            p[0] = "%s:%s" % (p[1], p[2])

        def p_error(self, p):
            print("Syntax Error: %s" % (p,))
            self._parser.restart()

        def build(self, **kargs):
            opt = dict(debug=0, write_tables=0)
            opt.update(**kargs)
            if _has_ply:
                self._parser = yacc.yacc(module=self, **opt)

    def __init__(self, **kargs):
        self.lexer = Dot.Lexer()
        self.parser = Dot.Parser()
        if not _has_ply:
            print("warning: Dot parser not supported (install python-ply)")

    def parse(self, data):
        try:
            self.parser._parser.restart()
        except AttributeError:
            self.lexer.build(reflags=lex.re.UNICODE)
            self.parser.build()
        except:
            print("unexpected error")
            return None
        try:
            s = data.decode("utf-8")
        except UnicodeDecodeError:
            s = data
        L = self.parser._parser.parse(s, lexer=self.lexer._lexer)
        return L

    def read(self, filename):
        f = file(
            filename, "rb"
        )  # As it'll try to decode later on with utf-8, read it binary at this point.
        return self.parse(f.read())


class Point(object):
    """
    Helper class representing a point.
    """

    def __init__(self, *pts):
        self.x, self.y = pts

    def __getitem__(self, i):
        if i == 0:
            return self.x

        if i == 1:
            return self.y

        raise AssertionError(
            "For 2d point can only get 0 or 1 (trying to get: %s)" % (i,)
        )

    def __len__(self):
        return 2

    def __iter__(self):
        yield self.x
        yield self.y

    def distance(self, p2):
        x1, y1 = self
        x2, y2 = p2
        dist = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        return dist

    def __str__(self):
        return "(%s, %s)" % (self.x, self.y)

    def __repr__(self):
        return "Point(%s, %s)" % (self.x, self.y)


def angle_between_360_degrees(angle):
    while angle >= 360.0:
        angle -= 360.0
    while angle < 0.0:
        angle += 360.0
    if angle == 360.0:
        angle = 0.0

    return angle


def angle_to_x_axis_in_degrees(pt0, pt1):
    base_p0 = (0.0, 0.0)
    base_p1 = (1.0, 0.0)

    line1 = (pt0[0] - pt1[0], pt0[1] - pt1[1])
    line2 = (base_p1[0] - base_p0[0], base_p1[1] - base_p0[1])

    x1 = float(line1[0])
    y1 = float(line1[1])
    x2 = float(line2[0])
    y2 = float(line2[1])
    divide_by = abs(x1 * x2) + abs(y1 * y2)

    if divide_by == 0.0:
        if pt0[1] > pt1[1]:
            return 90.0
        else:
            return 270.0
    else:
        tg = ((x1 * y2) - (x2 * y1)) / divide_by
        tan_degrees = math.degrees(math.atan(tg))

        if pt0[0] > pt1[0]:
            tan_degrees = -tan_degrees
        else:
            tan_degrees += 180.0

        angle = tan_degrees
        angle = angle_between_360_degrees(angle)

        return angle


def new_point_at_distance(pt, distance, angle):
    angle = float(angle)
    x, y = pt[0], pt[1]
    x += float(distance) * numpy.cos(numpy.deg2rad(angle))
    y += float(distance) * numpy.sin(numpy.deg2rad(angle))
    return float(x), float(y)



from __future__ import print_function

import atexit
import json
import socket
import threading
import time
import uuid
import warnings
import weakref

try:
    from http.client import BadStatusLine
    from urllib.request import urlopen
except ImportError:
    from httplib import BadStatusLine
    from urllib2 import urlopen

import nengo_gui
from IPython import get_ipython
from IPython.display import HTML, display
from notebook.base.handlers import IPythonHandler
from notebook.utils import url_path_join
from tornado import gen, httpclient, httputil
from tornado.simple_httpclient import SimpleAsyncHTTPClient
from tornado.websocket import WebSocketHandler, websocket_connect


class ConfigReuseWarning(UserWarning):
    pass


class InlineGUI(object):
    shutdown_hook_registered = False

    servers = weakref.WeakValueDictionary()
    configs = set()

    host = "localhost"

    def __init__(self, model, cfg=None, height=600, backend="nengo"):
        self.height = height

        self._started = False

        if cfg is None:
            cfg = get_ipython().mktempfile()

        self.server = self.start_server(cfg, model)
        self.port = self.server.server.server_port
        self.server.server.settings.prefix = None
        self.server.server.page_settings.backend = backend

        self.resource = url_path_join(
            "/nengo", str(self.port), str(self.server.server.get_resource())
        )

    @classmethod
    def start_server(cls, cfg, model):
        if not cls.shutdown_hook_registered:
            atexit.register(InlineGUI.shutdown_all, timeout=5)
            cls.shutdown_hook_registered = True

        # Make sure only one server is writing the same config.
        server = cls.servers.get(cfg, None)
        if server is not None and server.is_alive():
            warnings.warn(
                ConfigReuseWarning(
                    "Reusing config. Only the most recent visualization will "
                    "update the config."
                )
            )
            for page in server.server.pages:
                page.save_config(force=True)
                page.filename_cfg = get_ipython().mktempfile()
                cls.servers[page.filename_cfg] = server

        name = model.label
        server_settings = nengo_gui.guibackend.GuiServerSettings(("localhost", 0))
        model_context = nengo_gui.guibackend.ModelContext(
            model=model, locals=get_ipython().user_ns, filename=None, writeable=False
        )
        page_settings = nengo_gui.page.PageSettings(
            filename_cfg=cfg, editor_class=nengo_gui.components.editor.NoEditor
        )
        server = nengo_gui.gui.GuiThread(model_context, server_settings, page_settings)
        server.start()
        cls.servers[cfg] = server
        cls.configs.add(cfg)
        return server

    @classmethod
    def shutdown_all(cls, timeout=None):
        def get_timeout(start=time.clock()):
            if timeout is None:
                return None
            else:
                return max(0.001, timeout - (time.clock() - start))

        for cfg in cls.configs:
            server = cls.servers.get(cfg, None)
            server.shutdown(get_timeout())

    def _ipython_display_(self):
        display(
            HTML(
                r"""
            <script type="text/javascript" id="{uuid}">
            {{
                let req = new XMLHttpRequest();
                req.addEventListener("load", function() {{
                    if (this.status != 200 && this.response != 'OK') {{
                        let p = document.getElementById('{uuid}').parentNode;
                        p.innerHTML +=
                            'The nengo_gui.jupyter notebook server ' +
                            'extension was not loaded. Please activate it ' +
                            'with the following command:' +
                            '<pre>jupyter serverextension enable ' +
                            'nengo_gui.jupyter</pre>';
                        p.classList.add('output_stderr');
                    }}
                }});
                req.open('GET', './nengo/check', true);
                req.send();
            }}
            </script>
        """.format(
                    uuid=uuid.uuid4()
                )
            )
        )

        self.server.wait_for_startup()
        if self.server.is_alive():
            vdom = {
                "tagName": "div",
                "attributes": {"id": str(uuid.uuid4())},
                "children": [
                    {
                        "tagName": "iframe",
                        "attributes": {
                            "src": "." + str(self.resource),
                            "width": "100%",
                            "height": str(self.height),
                            "frameborder": "0",
                            "class": "cell",
                            "allowfullscreen": "allowfullscreen",
                            "style": {
                                "border": "1px solid #eee",
                                "boxSizing": "border-box",
                            },
                        },
                    }
                ],
            }
            html = """
                <div id="{id}">
                    <iframe
                        src=".{url}"
                        width="100%"
                        height="{height}"
                        frameborder="0"
                        class="cell"
                        style="border: 1px solid #eee; box-sizing: border-box;"
                        allowfullscreen></iframe>
                </div>
            """.format(
                url=self.resource, id=uuid.uuid4(), height=self.height
            )
            bundle = {
                "application/vdom.v1+json": vdom,
                "text/html": html,
            }
            display(bundle, raw=True)
        else:
            print("Server is not alive.")


class IPythonViz(InlineGUI):
    def __init__(self, *args, **kwargs):
        warnings.warn(DeprecationWarning("IPythonViz has been renamed to InlineGUI."))
        super(IPythonViz, self).__init__(*args, **kwargs)


class LabServerManager(object):
    shutdown_hook_registered = False
    server = None

    @classmethod
    def start_server(cls, base_url):
        if not cls.shutdown_hook_registered:
            atexit.register(LabServerManager.shutdown, timeout=5)
            cls.shutdown_hook_registered = True

        if cls.server is not None:
            return cls.server

        server_settings = nengo_gui.guibackend.GuiServerSettings(
            listen_addr=("localhost", 0), auto_shutdown=0
        )
        model_context = nengo_gui.guibackend.ModelContext()

        cls.server = nengo_gui.gui.GuiThread(model_context, server_settings)
        cls.server.start()
        cls.server.server.settings.prefix = url_path_join(
            base_url, "nengo/" + str(cls.server.server.server_port)
        )
        cls.server.wait_for_startup()
        return cls.server

    @classmethod
    def shutdown(cls, timeout=None):
        if cls.server is not None:
            cls.server.shutdown(timeout)
            cls.server = None


class NengoGuiHandler(IPythonHandler):
    def __init__(self, *args, **kwargs):
        super(NengoGuiHandler, self).__init__(*args, **kwargs)
        self.status_code_read = False
        self.header_keys = set()

    @gen.coroutine
    def get(self, port):
        client = httpclient.AsyncHTTPClient()
        headers = httputil.HTTPHeaders()
        for name, value in self.request.headers.get_all():
            if name.lower() == "host":
                headers.add(name, "localhost:" + port)
            elif name.lower() == "origin":
                headers.add(name, "http://localhost:" + port)
            else:
                headers.add(name, value)
        request = httpclient.HTTPRequest(
            "http://localhost:" + port + self.request.uri,
            headers=headers,
            header_callback=self.header_callback,
            streaming_callback=self.streaming_callback,
        )
        response = yield client.fetch(request)
        if response.error is not None:
            response.rethrow()
        else:
            self.finish()

    def compute_etag(self):
        return None

    def header_callback(self, data):
        if not self.status_code_read:
            status = httputil.parse_response_start_line(data)
            self.set_status(status.code, status.reason)
            self.status_code_read = True
        else:
            for name, value in httputil.HTTPHeaders.parse(data).get_all():
                if name in self.header_keys:
                    self.add_header(name, value)
                else:
                    self.set_header(name, value)
                    self.header_keys.add(name)

    def streaming_callback(self, data):
        self.write(data)


class NengoGuiWSHandler(IPythonHandler, WebSocketHandler):
    def open(self, port, query_str):
        self.ws = None
        url = "ws://localhost:" + port + self.request.uri
        headers = httputil.HTTPHeaders()
        for name, value in self.request.headers.get_all():
            if name.lower() == "host":
                headers.add(name, "localhost:" + port)
            elif name.lower() == "origin":
                headers.add(name, "http://localhost:" + port)
            else:
                headers.add(name, value)
        request = httpclient.HTTPRequest(url, headers=headers)
        websocket_connect(
            request,
            callback=self.conn_callback,
            on_message_callback=self.on_message_callback,
        )

    def conn_callback(self, ws):
        self.ws = ws.result()

    def on_message(self, message):
        if isinstance(message, bytes):
            self.ws.write_message(message, binary=True)
        else:
            self.ws.write_message(message)

    def on_close(self):
        self.ws.close(self.close_code, self.close_reason)

    def on_message_callback(self, message):
        if message is None:
            self.close()
        elif isinstance(message, bytes):
            self.write_message(message, binary=True)
        else:
            self.write_message(message)


class AvailabilityCheckHandler(IPythonHandler):
    def get(self):
        self.finish("OK")


class StartGuiHandler(IPythonHandler):
    def set_default_headers(self):
        self.set_header("Content-Type", "application/json")

    def get(self):
        server = LabServerManager.start_server(self.base_url)
        self.finish(
            json.dumps(
                {"port": server.server.server_port, "token": server.server.auth_token}
            )
        )


class RedirectHandler(IPythonHandler):
    def get(self, port, resource):
        self.redirect(
            url_path_join(self.base_url, "nengo", port, resource)
            + "/?"
            + self.request.query,
            permanent=True,
        )


def _jupyter_server_extension_paths():
    return []


def load_jupyter_server_extension(nb_server_app):
    web_app = nb_server_app.web_app
    host_pattern = ".*$"
    availability_check_pattern = url_path_join(
        web_app.settings["base_url"], ".*/nengo/check"
    )
    start_gui_pattern = url_path_join(web_app.settings["base_url"], "/nengo/start_gui")
    ws_route_pattern = url_path_join(
        web_app.settings["base_url"], "/nengo/(\\d+)/viz_component(\\?.*)?$"
    )
    route_pattern = url_path_join(web_app.settings["base_url"], "/nengo/(\\d+)/.*$")
    redirect_pattern = url_path_join(
        web_app.settings["base_url"], ".+/nengo/(\\d+)/(.*)$"
    )
    web_app.add_handlers(
        host_pattern,
        [
            (availability_check_pattern, AvailabilityCheckHandler),
            (start_gui_pattern, StartGuiHandler),
            (ws_route_pattern, NengoGuiWSHandler),
            (route_pattern, NengoGuiHandler),
            (redirect_pattern, RedirectHandler),
        ],
    )



# Question Answering with Control using SPA

# Now we will build this model again, using the spa (semantic pointer
# architecture) package built into Nengo 2.0.

# Press the play button to run the simulation.
# The top graph is the input to the visual subnet. When this input is a
# STATEMENT, there is no response shown in the motor graph and the input is
# stored in memory (shown in the memory graph).  To see bound pairs (e.g.
# RED*CIRCLE) in memory, you need to right-click on the memory graph and
# select 'show pairs'.  This can be cluttered but is more informative.
# When the input to 'visual' is a QUESTION, the motor graph shows the
# appropriate answer. For instance, when the input to visual is
# QUESTION+BLUE (showin in the visual graphs), the output from motor is SQUARE.
# Note this simulation is fairly sensitive to the dimensionality, you may
# need multiple runs or a higher dimension to get the expected results.

# Setup the environment
import nengo
import nengo.spa as spa
import numpy as np
from nengo.spa import Vocabulary

dim = 32  # The dimensionality of the vectors
rng = np.random.RandomState(11)
vocab = Vocabulary(dimensions=dim, rng=rng, max_similarity=0.1)

# Adding semantic pointers to the vocabulary
CIRCLE = vocab.parse("CIRCLE")
BLUE = vocab.parse("BLUE")
RED = vocab.parse("RED")
SQUARE = vocab.parse("SQUARE")
ZERO = vocab.add("ZERO", [0] * dim)

# Create the spa.SPA network to which we can add SPA objects
model = spa.SPA(label="Question Answering with Control", vocabs=[vocab])
with model:
    model.visual = spa.State(dim)
    model.motor = spa.State(dim)
    model.memory = spa.State(dim, feedback=1, feedback_synapse=0.1)

    actions = spa.Actions(
        "dot(visual, STATEMENT) --> memory=visual",
        "dot(visual, QUESTION) --> motor = memory * ~visual",
    )

    model.bg = spa.BasalGanglia(actions)
    model.thalamus = spa.Thalamus(model.bg)

    # Function for providing visual input
    def visual_input(t):
        if 0.1 < t < 0.3:
            return "STATEMENT+RED*CIRCLE"
        elif 0.35 < t < 0.5:
            return "STATEMENT+BLUE*SQUARE"
        elif 0.55 < t < 0.7:
            return "QUESTION+BLUE"
        elif 0.75 < t < 0.9:
            return "QUESTION+CIRCLE"
        return "ZERO"

    # Inputs
    model.input = spa.Input(visual=visual_input)



import os

import nengo
import nengo_gui
import numpy as np


def test_lorenz():
    pyfile = os.path.join(nengo_gui.__path__[0], "examples", "tutorial", "15-lorenz.py")

    model_objects = {}
    nengo.utils.stdlib.execfile(pyfile, model_objects)

    with model_objects["model"]:
        state_p = nengo.Probe(model_objects["x"], synapse=0.01)

    with nengo.Simulator(model_objects["model"]) as sim:
        sim.run(5)

    variance = np.var(sim.data[state_p], axis=0)
    assert np.all(variance > 40)



import time

import pytest
from nengo_gui import testing_tools as tt

test_files = tt.folder_location("examples/tutorial")
ignore = ["spa", "networks"]
test_files = [f for f in test_files if not any(i in f for i in ignore)]
# Gets rid of SPA and Network files


@pytest.mark.parametrize("test_file", test_files)
def test_tutorial_basic(driver, test_file):
    # Tests the first 18 tutorials. These are the tutorials that do not
    # utilize networks or SPA which require a different format of test

    # Test page response by clicking the reset button and applying
    # new code to ace-editor
    tt.reset_page(driver)
    time.sleep(1)
    tt.update_editor(driver, test_file)
    tt.mouse_scroll(driver, 500)
    time.sleep(2)
    side_script = (
        "var right = document.getElementById('rightpane');\n"
        "right.style.width = '200px';\n"
    )
    driver.execute_script(side_script)

    node_objects = driver.find_elements_by_xpath('//*[@class="node"]')
    ens_objects = driver.find_elements_by_xpath('//*[@class="ens"]')

    node_number = test_file.count("nengo.Node")
    ens_number = test_file.count("nengo.Ensemble")

    # Makes sure the correct number of ensembles and nodes were rendered
    assert len(node_objects) / 2 == node_number
    assert len(ens_objects) / 2 == ens_number

    # Tests whether the file compiles and runs
    hang_time = 25  # alloted time until test fails

    time_script = "var time = $('#ticks_tr');\n" "return time.find('td').text()\n"
    tt.start_stop_sim(driver)
    time_start = time.time()
    while time.time() - time_start < hang_time:
        sim_time = driver.execute_script(time_script)
        if float(sim_time) > 0:
            break
        time.sleep(0.1)
    else:
        assert False, "Did not compile in %f seconds" % hang_time



# A Single Neuron Model

# All models in nengo are built inside "networks". Inside a network, you
# can put more networks, and you can connect networks to each other. You can
# also put other objects such as neural populations or "ensembles" (which
# have individual neurons inside of them) inside the networks. For this model,
# you will make a network with one neuron.

# This model has parameters as described in the book, ensuring that it is an
# "on" neuron. The neuron will be slightly different each time you run this
# script, as many parameters are randomly chosen.

# Press the play button to run the simulation.
# The graph on the top left shows the input signal and the graph on the top
# right shows the value represented by the neuron. The filtered spikes from the
# neuron are shown in the graph on the bottom right.


import nengo

# Setup the environment
import numpy as np

# Create the network object to which we can add ensembles, connections, etc.
model = nengo.Network(label="A Single Neuron")

with model:
    # Input
    stim = nengo.Node(lambda t: np.cos(16 * t), label="input")

    # Ensemble with one neuron
    neuron = nengo.Ensemble(
        1, dimensions=1, encoders=[[1]]  # Represent a scalar
    )  # Sets the neurons firing rate to increase for positive input

    # Connecting input to ensemble
    nengo.Connection(stim, neuron)



# Addition

# In this model, you will see a transformation which is the basic property of
# single neurons (i.e., addition). Addition transforms two inputs into a single
# output which is their sum. You will construct a network that adds two inputs.
# The network utilizes two communication channels going into the same neural
# population. Addition is somewhat "free", since the incoming currents from
# different synaptic connections interact linearly.

# This model has ensembles A and B which represent the two inputs to be added.
# The 'Sum' ensemble represents the added value. All the parameters used in the
# model are as described in the book, with the sum ensemble having a radius of
# 2 to account for the maximum range of summing the input values.

# While connecting the inputs to the ensembles A and B, the transform is
# set to 1 (which is the default value) since this should be a communication
# channel. However as described in the book, you can scale a represented
# variable by a constant value by changing the transform. Example: if you
# set the transform of ensemble B to 0 and ensemble A to 2
# (i.e., nengo.Connection(stim_A, ens_A, transform=[2]) ), the sum will be twice
# of the stim_A. You will also need to set an appropriate radius for the
# Sum ensemble to avoid saturation when you change the transform values.

# Press the play button to run the simulation.
# The stim_A and stim_B graphs show the inputs to ensembles A and B
# respectively. The graphs A and B show the decoded value of the activity of
# ensembles A and B respectively. The sum graph shows that the decoded value of
# the activity in the Sum ensemble provides a good estimate of the sum of inputs
# A and B. You can use the sliders to change the input values provided by the
# stim_A and stim_B nodes.


# Setup the environment
import nengo
from nengo.dists import Uniform
from nengo.processes import Piecewise

# Create the network object to which we can add ensembles, connections, etc.
model = nengo.Network(label="Scalar Addition")

with model:
    # Inputs to drive the activity in ensembles A and B
    stim_A = nengo.Node(
        Piecewise({0: -0.75, 1.25: 0.5, 2.5: 0.70, 3.75: 0}), label="Input A"
    )
    stim_B = nengo.Node(
        Piecewise({0: 0.25, 1.25: -0.5, 2.5: 0.85, 3.75: 0}), label="Input B"
    )

    # Ensembles with 100 LIF neurons each
    # Represents the first input (A)
    ens_A = nengo.Ensemble(100, dimensions=1, max_rates=Uniform(100, 200), label="A")
    # Represents the second input (B)
    ens_B = nengo.Ensemble(100, dimensions=1, max_rates=Uniform(100, 200), label="B")
    # Reprsents the sum of two inputs
    ens_sum = nengo.Ensemble(
        100, dimensions=1, max_rates=Uniform(100, 200), radius=2, label="Sum"
    )

    # Connecting the input nodes to ensembles
    nengo.Connection(stim_A, ens_A)
    nengo.Connection(stim_B, ens_B)

    # Connecting ensembles A and B to the Sum ensemble
    nengo.Connection(ens_A, ens_sum)
    nengo.Connection(ens_B, ens_sum)



# Nonlinear Transformations

# The real world is filled with non-linearities and so dealing
# with them often requires nonlinear computation. This model shows
# how to compute nonlinear functions using Nengo 2.0. The two nonlinear
# functions demonstrated in this model are 'multiplication' and 'squaring'.

# The parameters of the model are as described in the book. The model has five
# ensembles: two input ensembles (ens_X and ens_Y), a 2D combined ensemble (vector2D),
# and the output ensembles result_square and result_product which store the
# square and product of the inputs respectively.

# Two varying scalar values are used for the two input signals that drive
# activity in ensembles A and B. For multiplication, we project both
# inputs independently into a 2D space, and then decode a nonlinear
# transformation of that space (i.e., the product) into an ensemble
# (result_product). The model also squares the value of the first input
# (input_X) encoded in an ensemble (ens_X), in the output of another ensemble
# (result_square).

# The two functions product(x) and square(x) are defined to serve the same
# purpose as entering the expressions in the "Expression" field in the
# "User-defined Function" dialog box in Nengo 1.4 as described in the book.

# Press the play button to run the simulation.
# The input signals chosen clearly show that the model works well.
# The result_product graph shows the product of the input_X & input_Y, and
# the result_square graph shows the squre of input_X. You can see in the graphs
# that when input_X is zero, both the product and the square are also zero. You
# can use the sliders to change the input values provided by the input_X and
# input_Y nodes to test the model.

# To inspect any values in the graphs, you can hover your mouse over the graph
# and the x and t values will be displayed near the axis.  This can be easier
# when you pause the simulation.  When you pause, you can also drag the blue
# box at the bottom of the screen to slide backwards and forward through time.

# Setup the environment
import nengo
from nengo.processes import Piecewise

# Create the network object to which we can add ensembles, connections, etc.
model = nengo.Network(label="Nonlinear Function")

with model:
    # Input - Piecewise step functions
    input_X = nengo.Node(
        Piecewise({0: -0.75, 1.25: 0.5, 2.5: -0.75, 3.75: 0}), label="Input X"
    )
    input_Y = nengo.Node(
        Piecewise({0: 1, 1.25: 0.25, 2.5: -0.25, 3.75: 0.75}), label="Input Y"
    )

    # Five ensembles containing LIF neurons
    # Represents input_X
    ens_X = nengo.Ensemble(100, dimensions=1, radius=1, label="X")
    # Represents input_Y
    ens_Y = nengo.Ensemble(100, dimensions=1, radius=1, label="Y")
    # 2D ensemble to represent combined X and Y values
    vector2D = nengo.Ensemble(224, dimensions=2, radius=2)

    # Represents the square of X
    result_square = nengo.Ensemble(100, dimensions=1, radius=1, label="Square")
    # Represents the product of X and Y
    result_product = nengo.Ensemble(100, dimensions=1, radius=1, label="Product")

    # Connecting the input nodes to the appropriate ensembles
    nengo.Connection(input_X, ens_X)
    nengo.Connection(input_Y, ens_Y)

    # Connecting input ensembles A and B to the 2D combined ensemble
    nengo.Connection(ens_X, vector2D[0])
    nengo.Connection(ens_Y, vector2D[1])

    # Defining a function that computes the product of two inputs
    def product(x):
        return x[0] * x[1]

    # Defining the squaring function
    def square(x):
        return x[0] * x[0]

    # Connecting the 2D combined ensemble to the result ensemble
    nengo.Connection(vector2D, result_product, function=product)

    # Connecting ensemble A to the result ensemble
    nengo.Connection(ens_X, result_square, function=square)



try:
    from urllib.parse import quote, urlencode
except ImportError:
    from urllib import quote, urlencode


class Resource(object):
    __slots__ = ["path", "query"]

    def __init__(self, path=None, query=None):
        self.path = path
        self.query = query

    def __str__(self):
        url = []
        if self.path is not None:
            url.append(quote(self.path))
        if self.query is not None:
            url.extend(("?", urlencode(self.query)))
        return "".join(url)


class URL(object):
    __slots__ = ["protocol", "host", "port", "resource"]

    def __init__(self, host, port, protocol="http", resource=None):
        self.host = host
        self.port = port
        self.protocol = protocol
        self.resource = resource

    def __str__(self):
        url = []
        if self.protocol is not None:
            url.append(self.protocol)
        url.extend(("://", self.host, ":", str(self.port)))
        if self.resource is not None:
            url.append(str(self.resource))
        return "".join(url)



import nengo
import numpy as np

model = nengo.Network()
with model:
    stimulus_A = nengo.Node([1], label="stim A")
    stimulus_B = nengo.Node(lambda t: np.sin(2 * np.pi * t))
    ens = nengo.Ensemble(n_neurons=1000, dimensions=2)
    result = nengo.Ensemble(n_neurons=50, dimensions=1)
    nengo.Connection(stimulus_A, ens[0])
    nengo.Connection(stimulus_B, ens[1])
    nengo.Connection(ens, result, function=lambda x: x[0] * x[1], synapse=0.01)

    with nengo.Network(label="subnet") as subnet:
        a = nengo.Ensemble(100, 1)
        b = nengo.Ensemble(100, 1)
        nengo.Connection(a, b)
        nengo.Connection(b, b)

        with nengo.Network() as subsubnet:
            c = nengo.Ensemble(100, 1)
            d = nengo.Ensemble(100, 1)
            nengo.Connection(c, d)
        nengo.Connection(b, c)
        nengo.Connection(d, a)
    nengo.Connection(result, a)


if __name__ == "__main__":
    import nengo_gui

    nengo_gui.GUI(__file__).start()



import nengo
import nengo.spa as spa

D = 16

model = spa.SPA(seed=1)
with model:
    model.a = spa.Buffer(dimensions=D)
    model.b = spa.Buffer(dimensions=D)
    model.c = spa.Buffer(dimensions=D)
    model.cortical = spa.Cortical(
        spa.Actions(
            "c = a+b",
        )
    )

    model.input = spa.Input(
        a="A",
        b=(lambda t: "C*~A" if (t % 0.1 < 0.05) else "D*~A"),
    )

if __name__ == "__main__":
    import nengo_gui

    nengo_gui.GUI(__file__).start()



import collections
import struct

import nengo
from nengo_gui.components.component import Component


class XYValue(Component):
    """Represents (at least) two dimensional values as co-ordinates on an
    x-y plot."""

    config_defaults = dict(
        max_value=1, min_value=-1, index_x=0, index_y=1, **Component.config_defaults
    )

    def __init__(self, obj):
        super(XYValue, self).__init__()
        self.obj = obj
        self.data = collections.deque()
        self.n_lines = int(obj.size_out)
        self.struct = struct.Struct("<%df" % (1 + self.n_lines))
        self.node = None
        self.conn = None

    def attach(self, page, config, uid):
        super(XYValue, self).attach(page, config, uid)
        self.label = page.get_label(self.obj)

    def add_nengo_objects(self, page):
        with page.model:
            self.node = nengo.Node(self.gather_data, size_in=self.obj.size_out)
            self.conn = nengo.Connection(self.obj, self.node, synapse=0.01)

    def remove_nengo_objects(self, page):
        page.model.connections.remove(self.conn)
        page.model.nodes.remove(self.node)

    def gather_data(self, t, x):
        self.data.append(self.struct.pack(t, *x))

    def update_client(self, client):
        while len(self.data) > 0:
            data = self.data.popleft()
            client.write_binary(data)

    def javascript(self):
        info = dict(uid=id(self), n_lines=self.n_lines, label=self.label)
        json = self.javascript_config(info)
        return "new Nengo.XYValue(main, sim, %s);" % json

    def code_python_args(self, uids):
        return [uids[self.obj]]



# Learning a communication channel

# This model shows how to include synaptic plasticity in Nengo models using the
# hPES learning rule. Here we implement error-driven learning to learn to
# compute a simple communication channel. This is done by using an error signal
# provided by a neural ensemble to modulate the connection between two other
# ensembles.

# The model has parameters as described in the book. Note that hPES learning
# rule is built into Nengo 1.4 as mentioned in the book. The same rule can be
# implemented in Nengo 2.0 by combining the PES and the BCM rule as shown in the
# code. Also, instead of using the 'gate' and 'switch' as described in the book,
# an 'inhibit' population is used which serves the same purpose of turning off the
# learning by inhibiting the error population.

# Note that to compute the actual error value (which is required only for analysis),
# the book uses a population of "Direct" mode neurons. In Nengo 2.0, this can be done
# more efficiently using a 'nengo.Node()'.

# When you run the model, you will see that the 'post' population gradually
# learns to compute the communication channel. In the model, you will inhibit
# the error population after 15 seconds to turn off learning and you will see
# that the 'post' population will still track the 'pre' population showing that
# the model has actually learned the input.

# Press the play button to run the simulation.
# From the plots, you will see that the post population initially doesn't track
# the pre population but eventually it learns the function (which is a
# communication channel in this case) and starts tracking the post population.
# After the error is inhibited (i.e., t>15s, shown by value zero in the error
# plot), the post population continues to track the pre population. The
# actual_error graph shows that there is significant error between the pre and
# the post populations at the begining which eventually gets reduced to almost
# zero as model learns the communication channel.

# The model can also learn other functions by using an appropriate error signal.
# For example to learn a square function, comment out the lines marked
# "# Learn the communication channel" and uncomment the lines marked
# "# Learn the square function" in the code. Run the model again and you will see
# that the model successfully learns the square function.

import nengo

# Setup the environment
import numpy as np
from nengo.processes import WhiteSignal

# Create the network object to which we can add ensembles, connections, etc.
model = nengo.Network(label="Learning", seed=7)
with model:
    # Ensembles to represent populations
    pre = nengo.Ensemble(50, dimensions=1, label="Pre")
    post = nengo.Ensemble(50, dimensions=1, label="Post")

    # Ensemble to compute the learning error signal
    error = nengo.Ensemble(100, dimensions=1, label="Learning Error")

    # Node to compute the actual error value
    actual_error = nengo.Node(size_in=1, label="Actual Error")

    # Learn the communication channel
    nengo.Connection(pre, actual_error, transform=-1)
    nengo.Connection(pre, error, transform=-1, synapse=0.02)

    # Learn the square function
    # nengo.Connection(pre, actual_error, function=lambda x: x**2, transform=-1)
    # nengo.Connection(pre, error, function=lambda x: x**2, transform=-1)

    # Error = pre - post
    nengo.Connection(post, actual_error, transform=1)
    nengo.Connection(post, error, transform=1, synapse=0.02)

    # Connecting pre population to post population (communication channel)
    conn = nengo.Connection(
        pre,
        post,
        function=lambda x: np.random.random(1),
        solver=nengo.solvers.LstsqL2(weights=True),
    )

    # Adding the learning rule to the connection
    conn.learning_rule_type = {
        "my_pes": nengo.PES(),
        "my_bcm": nengo.BCM(learning_rate=1e-10),
    }

    # Error connections don't impart current
    error_conn = nengo.Connection(error, conn.learning_rule["my_pes"])

    # Providing input to the model
    stim = nengo.Node(WhiteSignal(30, high=10), label="Input")  # RMS = 0.5 by default
    nengo.Connection(stim, pre, synapse=0.02)  # Connect the input to the pre ensemble

    # Function to inhibit the error population after 15s
    def inhib(t):
        return 2.0 if t > 15.0 else 0.0

    # Connecting inhibit population to error population
    inhibit = nengo.Node(inhib, label="Inhibit")
    nengo.Connection(
        inhibit, error.neurons, transform=[[-3]] * error.n_neurons, synapse=0.01
    )



# Structured Representations using SPA

# Now we will build the model again, using the spa (semantic pointer
# architecture) package built into Nengo 2.0. You will see that using the spa
# package considerably simplifies model construction and visualization through
# nengo_gui.

# Press the play button to run the simulation.
# The graphs A and B show the semantic pointer representations in objects A and
# B respectively. Graphs labelled C show the result of the convolution operation
# (left - shows the semantic pointer representation in object C, right - shows
# the similarity with the vectors in the vocabulary). The graphs labelled Sum
# show the sum of A and B as represented by the object Sum (left - shows the
# semantic pointer representation in Sum, right - shows high similarity with
# vectors A and B).

# Setup the environment
import nengo
import nengo.spa as spa
import numpy as np
from nengo.spa import Vocabulary

dim = 32  # the dimensionality of the vectors

# Creating a vocabulary
rng = np.random.RandomState(0)
vocab = Vocabulary(dimensions=dim, rng=rng)
vocab.add("C", vocab.parse("A * B"))

# Create the spa.SPA network to which we can add SPA objects
model = spa.SPA(label="structure", vocabs=[vocab])
with model:
    model.A = spa.State(dim)
    model.B = spa.State(dim)
    model.C = spa.State(dim, feedback=1)
    model.Sum = spa.State(dim)

    actions = spa.Actions("C = A * B", "Sum = A", "Sum = B")

    model.cortical = spa.Cortical(actions)

    # Model input
    model.input = spa.Input(A="A", B="B")



"""Nengo GUI backend implementation."""

from __future__ import print_function

import json
import logging
import mimetypes
import os
import os.path
import pkgutil

try:
    from urllib.parse import unquote
except ImportError:  # Python 2.7
    from urllib import unquote
try:
    import ssl
except ImportError:  # for Python without ssl support
    from . import nossl as ssl

import time

import nengo_gui
import nengo_gui.exec_env
import nengo_gui.page
from nengo_gui import server, url
from nengo_gui._vendor.cookies import Cookie
from nengo_gui.completion import get_completions
from nengo_gui.password import checkpw, gensalt

logger = logging.getLogger(__name__)


class Session(object):
    __slots__ = ["creation_time", "authenticated", "peer_name", "login_host"]

    def __init__(self):
        self.creation_time = time.time()
        self.authenticated = False
        self.peer_name = None
        self.login_host = None


class SessionManager(object):
    def __init__(self, time_to_live):
        self.time_to_live = time_to_live
        self._sessions = {}

    def __getitem__(self, session_id):
        session = self._sessions.get(session_id, None)
        if session is None or session.creation_time + self.time_to_live < time.time():
            del self._sessions[session_id]
            raise KeyError("Session expired")
        return session

    def __len__(self):
        return len(self._sessions)

    def add_session(self, request, session):
        session_id = self._new_session_id(request)
        self._sessions[session_id] = session
        return session_id

    def _new_session_id(self, request):
        return gensalt(16)


class RequireAuthentication(object):
    def __init__(self, login_page):
        self.login_page = login_page

    @classmethod
    def get_token(cls, inst):
        if "token" in inst.db:
            return inst.db["token"]
        elif inst.headers.get("Authorization", "").lower().startswith("token "):
            return inst.headers.get("Authorization").split(" ")[1]
        return None

    def __call__(self, fn):
        def auth_checked(inst):
            session = inst.get_session()
            if session.authenticated:
                return fn(inst)
            elif inst.server.verify_token(self.get_token(inst)):
                self.authenticate(inst)
                return fn(inst)
            return server.HttpRedirect(inst.server.settings.prefix + self.login_page)

        return auth_checked

    @classmethod
    def authenticate(cls, request_handler):
        session = request_handler.get_session()
        session.authenticated = True
        session.login_host = request_handler.headers.get("host", None)
        try:
            session.peer_name = request_handler.request.getpeername()[0]
        except Exception:
            logger.warning(
                "Cannot get peer name. Session will not be tied to " "client.",
                exc_info=True,
            )
        request_handler.persist_session(session)


class GuiRequestHandler(server.HttpWsRequestHandler):
    http_commands = {
        "/": "serve_main",
        "/login": "login_page",
        "/static": "serve_static",
        "/browse": "browse",
        "/complete": "complete",
        "/shutdown": "request_shutdown",
        "/favicon.ico": "serve_favicon",
    }

    def get_expected_origins(self):
        login_host = self.get_session().login_host
        return [login_host] if login_host is not None else []

    def http_GET(self):
        if self.server.settings.prefix is None:
            if self.server.verify_token(RequireAuthentication.get_token(self)):
                prefix = self.resource
                if prefix.endswith("/"):
                    prefix = prefix[:-1]
                self.server.settings.prefix = prefix
                RequireAuthentication.authenticate(self)
            else:
                raise server.InternalServerError("Prefix not set.")

        if not self.resource.startswith(self.server.settings.prefix):
            raise server.InvalidResource(self.resource)
        self.resource = self.resource[len(self.server.settings.prefix) :]
        server.HttpWsRequestHandler.http_GET(self)

    def login_page(self):
        session = self.get_session()
        content = b""

        if session.authenticated:
            return server.HttpRedirect(self.server.settings.prefix + "/")

        if "pw" in self.db:
            valid_pw = self.server.settings.password_hash is not None and checkpw(
                self.db["pw"], self.server.settings.password_hash
            )
            valid_token = self.server.verify_token(self.db["pw"])
            if valid_pw or valid_token:
                RequireAuthentication.authenticate(self)
                return server.HttpRedirect(self.server.settings.prefix + "/")
            else:
                content += b"<p><strong>Invalid password. Try again."
                content += b"</strong></p>"
        else:
            content += b"<p>Please enter the password or the security token "
            content += b"shown by Nengo when it was started:</p>"

        return server.HtmlResponse(
            content
            + b"""
            <form method="POST"><p>
                <label for="pw">Password: </label>
                <input type="password" name="pw" />
                <input type="submit" value="Log in" />
            </p></form>
        """
        )

    def serve_static(self):
        """Handles http://host:port/static/* by returning pkg data"""
        assert self.resource[0] == "/"
        static_dir = "static" + os.sep
        fn = os.path.normpath(self.resource[1:])
        if os.path.commonprefix((static_dir, fn)) != static_dir:
            raise server.Forbidden()
        mimetype, encoding = mimetypes.guess_type(fn)
        data = pkgutil.get_data("nengo_gui", fn)
        return server.HttpResponse(data, mimetype)

    @RequireAuthentication("/login")
    def browse(self):
        r = [b'<ul class="jqueryFileTree" style="display: none;">']
        d = unquote(self.db["dir"])
        ex_tag = "//examples//"
        ex_html = b"<em>built-in examples</em>"
        if d == ".":
            r.append(
                b'<li class="directory collapsed examples_dir">'
                b'<a href="#" rel="'
                + ex_tag.encode("utf-8")
                + b'">'
                + ex_html
                + b"</a></li>"
            )
            path = "."
        elif d.startswith(ex_tag):
            path = os.path.join(nengo_gui.__path__[0], "examples", d[len(ex_tag) :])
        else:
            path = os.path.join(".", d)

        for f in sorted(os.listdir(path)):
            ff = os.path.join(path, f).encode("utf-8")
            if os.path.isdir(os.path.join(path, f)):
                f = f.encode("utf-8")
                r.append(
                    b'<li class="directory collapsed">'
                    b'<a href="#" rel="' + ff + b'/">' + f + b"</a></li>"
                )
            else:
                e = os.path.splitext(f)[1][1:]  # get .ext and remove dot
                if e == "py":
                    e = e.encode("utf-8")
                    f = f.encode("utf-8")
                    r.append(
                        b'<li class="file ext_' + e + b'">'
                        b'<a href="#" rel="' + ff + b'">' + f + b"</a></li>"
                    )
        r.append(b"</ul>")
        return server.HtmlResponse(b"".join(r))

    @RequireAuthentication("/login")
    def complete(self):
        completions = get_completions(
            self.db["code"],
            int(self.db["row"]) + 1,
            int(self.db["col"]),
            self.db["filename"],
        )
        return server.JsonResponse(
            [
                {
                    "name": c.name,
                    "value": c.name,
                    "score": 1,
                    "meta": c.type,
                }
                for c in completions
            ]
        )

    @RequireAuthentication("/login")
    def serve_main(self):
        if self.resource != "/":
            raise server.InvalidResource(self.resource)

        filename = self.query.get("filename", [None])[0]
        reset_cfg = self.query.get("reset", [False])[0]
        page = self.server.create_page(filename, reset_cfg=reset_cfg)

        # read the template for the main page
        html = pkgutil.get_data("nengo_gui", "templates/page.html")
        if isinstance(html, bytes):
            html = html.decode("utf-8")

        # fill in the javascript needed and return the complete page
        components = page.create_javascript()
        data = (html % dict(components=components)).encode("utf-8")
        return server.HttpResponse(data)

    @RequireAuthentication("/login")
    def request_shutdown(self):
        self.server.shutdown()

    def serve_favicon(self):
        self.resource = "/static/favicon.ico"
        return self.serve_static()

    @RequireAuthentication("/login")
    def ws_default(self):
        """Handles ws://host:port/viz_component with a websocket"""
        # figure out what component is being connected to

        gui = self.server
        uid = int(self.query["uid"][0])

        component = gui.component_uids[uid]
        while self.ws.state is server.WebSocket.ST_OPEN:
            try:
                if component.replace_with is not None:
                    component.finish()
                    component = component.replace_with

                # read all data coming from the component
                msg = self.ws.read_frame()
                while msg is not None:
                    if not self._handle_ws_msg(component, msg):
                        return
                    msg = self.ws.read_frame()

                # send data to the component
                component.update_client(self.ws)
                component.page.save_config(lazy=True)
                time.sleep(0.01)
            except server.SocketClosedError:
                # This error means the server has shut down
                component.page.save_config(lazy=False)  # Stop nicely
                break
            except:
                logger.exception("Error during websocket communication.")

        # After hot loop
        component.finish()

    def _handle_ws_msg(self, component, msg):
        """Handle websocket message. Returns True when further messages should
        be handled and false when no further messages should be processed."""
        if msg.data.startswith("config:"):
            return self._handle_config_msg(component, msg)
        elif msg.data.startswith("remove"):
            return self._handle_remove_msg(component, msg)
        else:
            try:
                component.message(msg.data)
                return True
            except:
                logging.exception("Error processing: %s", repr(msg.data))

    def _handle_config_msg(self, component, msg):
        cfg = json.loads(msg.data[7:])
        old_cfg = {}
        for k in component.config_defaults.keys():
            v = getattr(component.page.config[component], k)
            old_cfg[k] = v
        if not cfg == old_cfg:
            # Register config change to the undo stack
            component.page.config_change(component, cfg, old_cfg)
        for k, v in cfg.items():
            setattr(component.page.config[component], k, v)
        component.page.modified_config()
        return True

    def _handle_remove_msg(self, component, msg):
        if msg.data != "remove_undo":
            # Register graph removal to the undo stack
            component.page.remove_graph(component)
        component.page.remove_component(component)
        component.page.modified_config()
        return False

    def get_session(self):
        try:
            peer_name = self.request.getpeername()[0]
        except Exception:
            peer_name = None

        try:
            session_id = self.request_cookies[
                "_sid_" + str(self.server.server_port)
            ].value
            session = self.server.sessions[session_id]
            if session.peer_name != peer_name:
                logger.warning(
                    "Session peer name mismatch: %s, %s", peer_name, session.peer_name
                )
                raise KeyError()
        except KeyError:
            session = Session()
        return session

    def persist_session(self, session):
        sid = self.server.sessions.add_session(self.request, session)
        sid_name = "_sid_" + str(self.server.server_port)
        self.response_cookies.add(
            Cookie(
                name=sid_name,
                value=sid,
                path=self.server.settings.prefix + "/",
                httponly=True,
                version=1,
                max_age=self.server.settings.session_duration,
            )
        )

    def log_message(self, format, *args):
        logger.info(format, *args)


class ModelContext(object):
    """Provides context information to a model. This can include the locals
    dictionary, the filename and whether this model can (or is allowed) to be
    written to disk."""

    __slots__ = ["model", "filename", "locals", "writeable"]

    def __init__(self, model=None, locals=None, filename=None, writeable=True):
        self.filename = filename
        if self.filename is not None:
            try:
                self.filename = os.path.relpath(filename)
            except ValueError:
                # happens on Windows if filename is on a different
                # drive than the current directory
                self.filename = filename
            self.writeable = writeable
        else:
            self.writeable = False

        if model is None and locals is not None:
            model = locals.get("model", None)

        self.model = model
        self.locals = locals


class GuiServerSettings(object):
    __slots__ = [
        "listen_addr",
        "auto_shutdown",
        "password_hash",
        "ssl_cert",
        "ssl_key",
        "session_duration",
        "prefix",
    ]

    def __init__(
        self,
        listen_addr=("localhost", 8080),
        auto_shutdown=2,
        password_hash=None,
        ssl_cert=None,
        ssl_key=None,
        session_duration=60 * 60 * 24 * 30,
        prefix="",
    ):
        self.listen_addr = listen_addr
        self.auto_shutdown = auto_shutdown
        self.password_hash = password_hash
        self.ssl_cert = ssl_cert
        self.ssl_key = ssl_key
        self.session_duration = session_duration
        self.prefix = prefix

    @property
    def use_ssl(self):
        if self.ssl_cert is None and self.ssl_key is None:
            return False
        elif self.ssl_cert is not None and self.ssl_key is not None:
            return True
        else:
            raise ValueError("SSL needs certificate file and key file.")


class GuiServer(server.ManagedThreadHttpServer):
    def __init__(
        self,
        model_context,
        server_settings=GuiServerSettings(),
        page_settings=nengo_gui.page.PageSettings(),
    ):
        if nengo_gui.exec_env.is_executing():
            raise nengo_gui.exec_env.StartedGUIException()
        self.settings = server_settings

        server.ManagedThreadHttpServer.__init__(
            self, self.settings.listen_addr, GuiRequestHandler
        )
        if self.settings.use_ssl:
            for b in self.bindings:
                b.socket = ssl.wrap_socket(
                    b.socket,
                    certfile=self.settings.ssl_cert,
                    keyfile=self.settings.ssl_key,
                    server_side=True,
                )

        self.auth_token = gensalt(24)
        self._one_time_auth_tokens = set()
        self.sessions = SessionManager(self.settings.session_duration)

        # the list of running Pages
        self.pages = []

        # a mapping from uids to Components for all running Pages.
        # this is used to connect the websockets to the appropriate Component
        self.component_uids = {}

        self.model_context = model_context
        self.page_settings = page_settings

        self._last_access = time.time()

    def create_page(self, filename, reset_cfg=False):
        """Create a new Page with this configuration"""
        page = nengo_gui.page.Page(
            self, filename=filename, settings=self.page_settings, reset_cfg=reset_cfg
        )
        self.pages.append(page)
        return page

    def remove_page(self, page):
        self._last_access = time.time()
        page.close()
        self.pages.remove(page)
        if (
            not self._shutting_down
            and self.settings.auto_shutdown > 0
            and len(self.pages) <= 0
        ):
            time.sleep(self.settings.auto_shutdown)
            earliest_shutdown = self._last_access + self.settings.auto_shutdown
            if earliest_shutdown < time.time() and len(self.pages) <= 0:
                logging.info("No connections remaining to the nengo_gui server.")
                self.shutdown()

    def verify_token(self, token):
        if token in self._one_time_auth_tokens:
            self._one_time_auth_tokens.remove(token)
            return True
        elif token == self.auth_token:
            return True
        else:
            return False

    def gen_one_time_token(self):
        token = gensalt(24)
        self._one_time_auth_tokens.add(token)
        return token

    def get_resource(self, action=None, token=True):
        path = []
        if self.settings.prefix is not None:
            path.append(self.settings.prefix)
            if action is None:
                path.append("")
        if action is not None:
            path.append(action)

        if token == "one-time":
            query = {"token": self.gen_one_time_token()}
        elif token:
            query = {"token": self.auth_token}
        else:
            query = None
        return url.Resource("/".join(path), query)

    def get_url(self, action=None, token=True):
        protocol = "https" if self.settings.use_ssl else "http"
        return url.URL(
            self.server_name,
            self.server_port,
            protocol,
            self.get_resource(action, token),
        )



import collections
import json
import os
import threading
import time
import traceback

import nengo
import nengo_gui.layout
import nengo_gui.user_action
import numpy as np
from nengo_gui.compat import escape, iteritems
from nengo_gui.components.component import Component
from nengo_gui.components.slider import OverriddenOutput
from nengo_gui.components.value import Value
from nengo_gui.modal_js import infomodal


class NetGraph(Component):
    """Handles computations and communications for NetGraph on the JS side.

    Communicates to all NetGraph components for creation, deletion and
    manipulation.
    """

    config_defaults = {}
    configs = {}

    def __init__(self):
        # this component must be ordered before all the normal graphs (so that
        # other graphs are on top of the NetGraph), so its
        # order is between that of SimControl and the default (0)
        super(NetGraph, self).__init__(component_order=-5)

        # this lock ensures safety between check_for_reload() and update_code()
        self.code_lock = threading.Lock()
        self.new_code = None

        self.uids = {}
        self.parents = {}
        self.initialized_pan_and_zoom = False

    def attach(self, page, config, uid):
        super(NetGraph, self).attach(page, config, uid)
        self.layout = nengo_gui.layout.Layout(self.page.model)
        self.to_be_expanded = collections.deque([self.page.model])
        self.to_be_sent = collections.deque()

        self.networks_to_search = [self.page.model]

        try:
            self.last_modify_time = os.path.getmtime(self.page.filename)
        except OSError:
            self.last_modify_time = None
        except TypeError:  # happens if self.filename is None
            self.last_modify_time = None

        self.last_reload_check = time.time()

    def check_for_reload(self):
        if self.page.filename is not None:
            try:
                t = os.path.getmtime(self.page.filename)
                if self.last_modify_time is None or self.last_modify_time < t:
                    self.reload()
                    self.last_modify_time = t
            except OSError:
                pass

        with self.code_lock:
            new_code = self.new_code
            # the lock is in case update_code() is called between these lines
            self.new_code = None

        if new_code is not None:
            self.reload(code=new_code)

    def update_code(self, code):
        """Set new version of code to display."""
        with self.code_lock:
            self.new_code = code

    def reload(self, code=None):
        """Called when new code has been detected
        checks that the page is not currently being used
        and thus can be updated"""
        with self.page.lock:
            self._reload(code=code)

    def _reload(self, code=None):  # noqa: C901
        """Loads and executes the code, removing old items,
        updating changed items
        and adding new ones"""

        old_locals = self.page.last_good_locals
        old_default_labels = self.page.default_labels

        if code is None:
            with open(self.page.filename) as f:
                code = f.read()
            if self.page.code == code:
                # don't re-execute the identical code
                return
            else:
                # send the new code to the client
                self.page.editor.update_code(code)

        self.page.execute(code)

        if self.page.error is not None:
            return

        name_finder = nengo_gui.NameFinder(self.page.locals, self.page.model)

        self.networks_to_search = [self.page.model]
        self.parents = {}

        removed_uids = {}
        rebuilt_objects = []

        # for each item in the old model, find the matching new item
        # for Nodes, Ensembles, and Networks, this means to find the item
        # with the same uid.  For Connections, we don't really have a uid,
        # so we use the uids of the pre and post objects.
        for uid, old_item in iteritems(dict(self.uids)):
            try:
                new_item = eval(uid, self.page.locals)
            except:
                new_item = None

            # check to make sure the new item's uid is the same as the
            # old item.  This is to catch situations where an old uid
            # happens to still refer to something in the new model, but that's
            # not the normal uid for that item.  For example, the uid
            # "ensembles[0]" might still refer to something even after that
            # ensemble is removed.
            new_uid = self.page.get_uid(new_item, default_labels=name_finder.known_name)
            if new_uid != uid:
                new_item = None

            same_class = False
            for cls in [nengo.Ensemble, nengo.Node, nengo.Network, nengo.Connection]:
                if isinstance(new_item, cls) and isinstance(old_item, cls):
                    same_class = True
                    break

            # find reasons to delete the object.  Any deleted object will
            # be recreated, so try to keep this to a minimum
            keep_object = True
            if new_item is None:
                keep_object = False
            elif not same_class:
                # don't allow changing classes
                keep_object = False
            elif self.get_extra_info(new_item) != self.get_extra_info(old_item):
                keep_object = False

            if not keep_object:
                self.to_be_sent.append(dict(type="remove", uid=uid))
                del self.uids[uid]
                removed_uids[old_item] = uid
                rebuilt_objects.append(uid)
            else:
                # fix aspects of the item that may have changed
                if self._reload_update_item(uid, old_item, new_item, name_finder):
                    # something has changed about this object, so rebuild
                    # the components that use it
                    rebuilt_objects.append(uid)

                self.uids[uid] = new_item

        self.to_be_expanded.append(self.page.model)

        self.page.name_finder = name_finder
        self.page.default_labels = name_finder.known_name
        self.page.config = self.page.load_config()
        self.page.uid_prefix_counter = {}
        self.layout = nengo_gui.layout.Layout(self.page.model)
        self.page.code = code

        orphan_components = []
        rebuild_components = []

        # items that are shown in components, but not currently displayed
        #  in the NetGraph (i.e. things that are inside collapsed
        #  Networks, but whose values are being shown in a graph)
        collapsed_items = []

        # remove graphs no longer associated to NetgraphItems
        removed_items = list(removed_uids.values())
        for c in self.page.components[:]:
            for item in c.code_python_args(old_default_labels):
                if item not in self.uids.keys() and item not in collapsed_items:

                    # item is a python string that is an argument to the
                    # constructor for the Component.  So it could be 'a',
                    # 'model.ensembles[3]', 'True', or even 'target=a'.
                    # We need to evaluate this string in the context of the
                    # locals dictionary and see what object it refers to
                    # so we can determine whether to rebuild this component.
                    #
                    # The following lambda should do this, handling both
                    # the normal argument case and the keyword argument case.
                    safe_eval = "(lambda *a, **b: " "list(a) + list(b.values()))(%s)[0]"

                    # this Component depends on an item inside a collapsed
                    #  Network, so we need to check if that component has
                    #  changed or been removed
                    old_obj = eval(safe_eval % item, old_locals)

                    try:
                        new_obj = eval(safe_eval % item, self.page.locals)
                    except:
                        # the object this Component depends on no longer exists
                        new_obj = None

                    if new_obj is None:
                        removed_items.append(item)
                    elif not isinstance(new_obj, old_obj.__class__):
                        rebuilt_objects.append(item)
                    elif self.get_extra_info(new_obj) != self.get_extra_info(old_obj):
                        rebuilt_objects.append(item)

                    # add this to the list of collapsed items, so we
                    # don't recheck it if there's another Component that
                    # also depends on this
                    collapsed_items.append(item)

                if item in rebuilt_objects:
                    self.to_be_sent.append(
                        dict(
                            type="delete_graph", uid=c.original_id, notify_server=False
                        )
                    )
                    rebuild_components.append(c.uid)
                    self.page.components.remove(c)
                    break
            else:
                for item in c.code_python_args(old_default_labels):
                    if item in removed_items:
                        self.to_be_sent.append(
                            dict(
                                type="delete_graph",
                                uid=c.original_id,
                                notify_server=False,
                            )
                        )
                        orphan_components.append(c)
                        break

        components = []
        # the old names for the old components
        component_uids = [c.uid for c in self.page.components]

        for name, obj in list(self.page.locals.items()):
            if isinstance(obj, Component):
                # the object has been removed, so the Component should
                #  be removed as well
                if obj in orphan_components:
                    continue

                # this is a Component that was previously removed,
                #  but is still in the config file, or it has to be
                #  rebuilt, so let's recover it
                if name not in component_uids:
                    self.page.add_component(obj)
                    self.to_be_sent.append(dict(type="js", code=obj.javascript()))
                    components.append(obj)
                    continue

                # otherwise, find the corresponding old Component
                index = component_uids.index(name)
                old_component = self.page.components[index]
                if isinstance(
                    obj,
                    (
                        nengo_gui.components.SimControlTemplate,
                        nengo_gui.components.AceEditorTemplate,
                        nengo_gui.components.NetGraphTemplate,
                    ),
                ):
                    # just keep these ones
                    components.append(old_component)
                else:
                    # replace these components with the newly generated ones
                    try:
                        self.page.add_component(obj)
                        old_component.replace_with = obj
                        obj.original_id = old_component.original_id
                    except:
                        traceback.print_exc()
                        print("failed to recreate plot for %s" % obj)
                    components.append(obj)

        components.sort(key=lambda x: x.component_order)

        self.page.components = components

        # notifies SimControl to pause the simulation
        self.page.changed = True

    def _reload_update_item(
        self, uid, old_item, new_item, new_name_finder
    ):  # noqa: C901
        """Tell the client about changes to the item due to reload."""
        changed = False

        if isinstance(old_item, (nengo.Node, nengo.Ensemble, nengo.Network)):

            old_label = self.page.get_label(old_item)
            new_label = self.page.get_label(
                new_item, default_labels=new_name_finder.known_name
            )

            if old_label != new_label:
                self.to_be_sent.append(
                    dict(type="rename", uid=uid, name=escape(new_label))
                )
                changed = True
            if isinstance(old_item, nengo.Network):
                if self.page.config[old_item].expanded:
                    self.to_be_expanded.append(new_item)
                    changed = True

        elif isinstance(old_item, nengo.Connection):
            old_pre = NetGraph.connection_pre_obj(old_item)
            old_post = NetGraph.connection_post_obj(old_item)
            new_pre = NetGraph.connection_pre_obj(new_item)
            new_post = NetGraph.connection_post_obj(new_item)

            old_pre = self.page.get_uid(old_pre)
            old_post = self.page.get_uid(old_post)
            new_pre = self.page.get_uid(
                new_pre, default_labels=new_name_finder.known_name
            )
            new_post = self.page.get_uid(
                new_post, default_labels=new_name_finder.known_name
            )

            if new_pre != old_pre or new_post != old_post:
                # if the connection has changed, tell javascript
                pres, posts = self.get_connection_hierarchy(
                    new_item, default_labels=new_name_finder.known_name
                )
                self.to_be_sent.append(
                    dict(type="reconnect", uid=uid, pres=pres, posts=posts)
                )
                changed = True
        return changed

    def get_parents(self, uid, default_labels=None):
        """Get parent networks for a connection"""
        while uid not in self.parents:
            net = self.networks_to_search.pop(0)
            net_uid = self.page.get_uid(net, default_labels=default_labels)
            for n in net.nodes:
                n_uid = self.page.get_uid(n, default_labels=default_labels)
                self.parents[n_uid] = net_uid
            for e in net.ensembles:
                e_uid = self.page.get_uid(e, default_labels=default_labels)
                self.parents[e_uid] = net_uid
            for n in net.networks:
                n_uid = self.page.get_uid(n, default_labels=default_labels)
                self.parents[n_uid] = net_uid
                self.networks_to_search.append(n)
        parents = [uid]
        while parents[-1] in self.parents:
            parents.append(self.parents[parents[-1]])
        return parents

    def modified_config(self):
        self.page.modified_config()

    def update_client(self, client):
        now = time.time()
        if now > self.last_reload_check + 0.5:
            self.check_for_reload()
            self.last_reload_check = now

        if not self.initialized_pan_and_zoom:
            self.send_pan_and_zoom(client)
            self.initialized_pan_and_zoom = True

        while len(self.to_be_sent) > 0:
            info = self.to_be_sent.popleft()
            client.write_text(json.dumps(info))

        if len(self.to_be_expanded) > 0:
            with self.page.lock:
                network = self.to_be_expanded.popleft()
                self.expand_network(network, client)

    def javascript(self):
        return 'new Nengo.NetGraph(main, {uid:"%s"});' % id(self)

    def message(self, msg):
        try:
            info = json.loads(msg)
        except ValueError:
            print("invalid message", repr(msg))
            return
        action = info.get("act", None)
        undo = info.get("undo", None)
        event = info.get("event", None)
        if action is not None:
            del info["act"]
            if action in ("auto_expand", "auto_collapse"):
                getattr(self, "act_" + action[5:])(**info)
            elif action in ("pan", "zoom", "create_modal"):
                # These should not use the undo stack
                getattr(self, "act_" + action)(**info)
            else:
                try:
                    act = nengo_gui.user_action.create_action(action, self, **info)
                    self.page.undo_stack.append([act])
                    del self.page.redo_stack[:]
                except:
                    print("error processing message", repr(msg))
                    traceback.print_exc()
        elif undo is not None:
            if undo == "1":
                self.undo()
            else:
                self.redo()
        elif event is not None:
            self.handle_event(event, info)
        else:
            print("received message", msg)

    def undo(self):
        if self.page.undo_stack:
            action = self.page.undo_stack.pop()
            re = []
            for act in action:
                act.undo()
                re.insert(0, act)
            self.page.redo_stack.append(re)

    def redo(self):
        if self.page.redo_stack:
            action = self.page.redo_stack.pop()
            un = []
            for act in action:
                act.apply()
                un.insert(0, act)
            self.page.undo_stack.append(un)

    def act_expand(self, uid):
        net = self.uids[uid]
        self.to_be_expanded.append(net)
        self.page.config[net].expanded = True
        self.modified_config()

    def act_collapse(self, uid):
        net = self.uids[uid]
        self.page.config[net].expanded = False
        self.remove_uids(net)
        self.modified_config()

    def remove_uids(self, net):
        for items in [net.ensembles, net.networks, net.nodes, net.connections]:
            for item in items:
                uid = self.page.get_uid(item)
                if uid in self.uids:
                    del self.uids[uid]
        for n in net.networks:
            self.remove_uids(n)

    def act_pan(self, x, y):
        self.page.config[self.page.model].pos = x, y
        self.modified_config()

    def act_zoom(self, scale, x, y):
        self.page.config[self.page.model].size = scale, scale
        self.page.config[self.page.model].pos = x, y
        self.modified_config()

    def act_create_modal(self, uid, **info):
        js = infomodal(self, uid, **info)
        self.to_be_sent.append(dict(type="js", code=js))

    def expand_network(self, network, client):
        """Display an expanded network, including the root network"""

        if not self.page.config[network].has_layout:
            pos = self.layout.make_layout(network)
            for obj, layout in pos.items():
                self.page.config[obj].pos = layout["y"], layout["x"]
                self.page.config[obj].size = layout["h"] / 2, layout["w"] / 2
            self.page.config[network].has_layout = True

        if network is self.page.model:
            parent = None
        else:
            parent = self.page.get_uid(network)
        for ens in network.ensembles:
            self.create_object(client, ens, obj_type="ens", parent=parent)
        for node in network.nodes:
            self.create_object(client, node, obj_type="node", parent=parent)
        for net in network.networks:
            self.create_object(client, net, obj_type="net", parent=parent)
        for conn in network.connections:
            self.create_connection(client, conn, parent=parent)
        self.page.config[network].expanded = True

    def create_object(self, client, obj, obj_type, parent):
        """Send the JSON of the newly created objects to client-side"""
        uid = self.page.get_uid(obj)

        # if the uid already exists, then it's already been inserted in
        # the netgraph, so don't send anything
        if uid in self.uids:
            return

        self.uids[uid] = obj

        pos = self.page.config[obj].pos
        if pos is None:
            import random

            pos = random.uniform(0, 1), random.uniform(0, 1)
            self.page.config[obj].pos = pos

        size = self.page.config[obj].size
        if size is None:
            size = (0.1, 0.1)
            self.page.config[obj].size = size

        label = self.page.get_label(obj)

        info = dict(
            uid=uid,
            label=escape(label),
            pos=pos,
            type=obj_type,
            size=size,
            parent=parent,
        )
        info.update(self.get_extra_info(obj))

        if type == "net":
            info["expanded"] = self.page.config[obj].expanded

        client.write_text(json.dumps(info))

    def get_extra_info(self, obj):
        """Determine helper information for each nengo object.

        This is used by the client side to configure the display.  It is also
        used by the reload() code to determine if a NetGraph object should
        be recreated.
        """
        info = {}
        if isinstance(obj, nengo.Node):
            if obj.output is None or (
                isinstance(obj.output, OverriddenOutput)
                and obj.output.base_output is None
            ):
                info["passthrough"] = True
            if callable(obj.output) and hasattr(obj.output, "_nengo_html_"):
                info["html"] = True
            info["dimensions"] = int(obj.size_out)
        elif isinstance(obj, nengo.Connection):
            info["kind"] = NetGraph.connection_kind(obj)
        elif isinstance(obj, nengo.Ensemble):
            info["dimensions"] = int(obj.size_out)
            info["n_neurons"] = int(obj.n_neurons)
        elif Value.default_output(obj) is not None:
            info["default_output"] = True

        info["sp_targets"] = nengo_gui.components.spa_plot.SpaPlot.applicable_targets(
            obj
        )
        return info

    def send_pan_and_zoom(self, client):
        pan = self.page.config[self.page.model].pos
        if pan is None:
            pan = 0, 0
        zoom = self.page.config[self.page.model].size
        if zoom is None:
            zoom = 1.0
        else:
            zoom = zoom[0]
        client.write_text(json.dumps(dict(type="pan", pan=pan)))
        client.write_text(json.dumps(dict(type="zoom", zoom=zoom)))

    @staticmethod
    def connection_pre_obj(conn):
        """
        Returns the pre-object of the given connection. Tries to make sure that
        the pre-object is an Ensemble. For example, if the connection originates
        from the ".neurons" object, the ensemble corresponding to the raw
        neurons is returned.
        """
        pre = conn.pre_obj
        if isinstance(pre, nengo.ensemble.Neurons):
            pre = pre.ensemble
        return pre

    @staticmethod
    def connection_post_obj(conn):
        """
        Returns the post-object relevant to the visualisation of the given
        connection. If the connection is a learning rule, returns the
        post-ensemble of that connection. If the connection ends in a
        ".neurons" object, returns the corresponding ensemble.
        """
        post = conn.post_obj
        if isinstance(post, nengo.connection.LearningRule):
            post = post.connection.post
            if isinstance(post, nengo.base.ObjView):
                post = post.obj
        if isinstance(post, nengo.ensemble.Neurons):
            post = post.ensemble
        return post

    @staticmethod
    def connection_kind(conn):
        """
        Categorises the given connection into one of three kinds: modulatory,
        inhibitory, and normal.
        """

        # Fetch the pre and post object for convenience
        pre, post = conn.pre_obj, conn.post_obj

        # Try to determine the connection kind by examining the connection class
        # and weight matrix
        kind = "normal"
        if hasattr(conn, "kind"):
            kind = conn.kind
        if isinstance(post, nengo.connection.LearningRule):
            kind = "modulatory"
        elif isinstance(post, nengo.ensemble.Neurons):
            trafo = conn.transform
            if hasattr(nengo, "transforms"):  # Support for Nengo 3.0
                if not isinstance(trafo, nengo.transforms.NoTransform):
                    trafo = trafo.sample()

            if hasattr(trafo, "size"):
                if trafo.size > 0 and (
                    np.all(trafo <= 0.0) and not np.all(np.isclose(trafo, 0.0))
                ):
                    return "inhibitory"

        # Support biologically plausible connections as e.g. provided by
        # nengo_bio
        if hasattr(conn, "dales_principle") and conn.dales_principle:
            if hasattr(pre, "p_exc") and (not pre.p_exc is None):
                is_purely_excitatory = np.allclose(pre.p_exc, 1.0)
                is_purely_inhibitory = np.allclose(pre.p_exc, 0.0)
                if kind == "inhibitory" and is_purely_excitatory:
                    kind = "dead"
                elif kind == "excitatory" and is_purely_inhibitory:
                    kind = "dead"

        return kind

    def get_connection_hierarchy(self, conn, default_labels=None):
        """
        For the given connection returns the pre and post objects and their
        hierarchy within the sub-network stack. Connection targets may either
        be ensembles or other connections (in the case of a connection
        targeting a learning rule.

        Parameters
        ----------
        conn : nengo.Connection
               Connection object for which the hierarchy should be computed.
        default_labels: nengo_gui.NameFinder
                        NameFinder instance used to derive the uids that are
                        being returned. If null, the default NameFinder instance
                        of this NetGraph instance is used.

        Returns
        -------
        (list, list)
            The first list contains the hierarchy for the pre-connection, the
            second list the hierarchy for the post-connection. The first element
            in either list corresponds to the actual target object, the
            remaining lists to the parent networks. All list elements are UIDs.
        """

        # Fetch the uid for the pre and post connection objects
        pre = self.page.get_uid(
            NetGraph.connection_pre_obj(conn), default_labels=default_labels
        )
        post = self.page.get_uid(
            NetGraph.connection_post_obj(conn), default_labels=default_labels
        )

        # Fetch the network hierarchy up to the post- and pre-connection object
        pres = self.get_parents(pre, default_labels=default_labels)[:-1]
        posts = self.get_parents(post, default_labels=default_labels)[:-1]

        # If this is a modulatory connection, connect to the connection that
        # is being trained
        if isinstance(conn.post_obj, nengo.connection.LearningRule):
            posts[0] = self.page.get_uid(
                conn.post_obj.connection, default_labels=default_labels
            )

        return pres, posts

    def create_connection(self, client, conn, parent):
        """
        Assembles the a JSON description of the given connection object and
        sends it to the client.

        Parameters
        ----------
        client : WebSocket
                 Websocket object to which the assembled JSON description is
                 sent.
        conn : nengo.Connection
               Connection object for which the description should be generated.
        parent : str
                 UID of the parent network the connection belongs to.
        """

        # Fetch the connection kind
        kind = kind = NetGraph.connection_kind(conn)
        if kind == "dead":
            return

        # Generate a uid for the connection
        uid = self.page.get_uid(conn)
        if uid in self.uids:
            return
        self.uids[uid] = conn

        # Fetch the pre and post object hierarchy
        pres, posts = self.get_connection_hierarchy(conn)

        # Serialise the connection descriptor and send it to the client
        info = dict(
            uid=uid, pre=pres, post=posts, type="conn", parent=parent, kind=kind
        )
        client.write_text(json.dumps(info))

    def handle_event(self, event, info):
        if event == "keyup":
            self.page.keys_pressed.discard(info["key"])
            self.page.key_codes_pressed.discard(info["keyCode"])
        elif event == "keydown":
            self.page.keys_pressed.add(info["key"])
            self.page.key_codes_pressed.add(info["keyCode"])



# Question Answering with Control

# This model shows a form of question answering where statements and
# questions are supplied through a single 'visual input' and the replies
# are produced in a 'motor output' as discussed in the book. You will
# implement this by using the basal ganglia to store and retrieve
# information from working memory in response to visual input. More
# specifically, the basal ganglia decides what to do with the information
# in the visual channel based on its content (i.e. whether it is a
# statement or a question).

# **Note: Simplified method of building the model using the spa (semantic
# pointer architecture) package in Nengo 2.0 is shown in the
# spa_question-control.py file in the same folder.

# This model has parameters as described in the book. Note that in Nengo 1.4,
# network arrays were used to construct this model for computational reasons
# as explained in the book. Nengo 2.0 has 'EnsembleArray' as an equivalent to
# network arrays which you will use in this model.

# When you run the model, it will start by binding RED and CIRCLE and then
# binding BLUE and SQUARE so the memory essentially has
# RED * CIRCLE + BLUE * SQUARE. This is stored in memory because the model is
# told that RED * CIRCLE is a STATEMENT (i.e. RED * CIRCLE + STATEMENT in the
# code) as is BLUE * SQUARE. Then it is presented with something like
# QUESTION + RED (i.e., "What is red?"). The basal ganglia then reroutes that
# input to be compared to what is in working memory and the result shows up in
# the motor channel.

# Press the play button to run the simulation.
# The graphs show that when the input to the Visual system is a STATEMENT,
# there is no response from the Motor system and the input is stored in the
# Memory. However, when the input to the Visual system is a QUESTION, the Motor
# system responds with the appropriate answer. For instance, when the input to
# Visual system is CIRCLE the output from the motor system is RED.


import nengo

# Setup the environment
import numpy as np
from nengo.dists import Uniform
from nengo.spa import Vocabulary

dim = 100  # Number of dimensions
n_neurons = 30  # Neurons per dimension
n_conv = 70  # Number of neurons per dimension in bind/unbind populations
n_mem = 50  # Number of neurons per dimension in memory population

# Defining a zero vector having length equal to the number of dimensions
ZERO = [0] * dim

# Creating the vocabulary
rng = np.random.RandomState(15)
vocab = Vocabulary(dimensions=dim, rng=rng, max_similarity=0.05)

# Create the network object to which we can add ensembles, connections, etc.
model = nengo.Network(label="Question Answering with Control", seed=15)
with model:
    # Ensembles
    visual = nengo.networks.EnsembleArray(
        n_neurons=n_neurons,
        n_ensembles=dim,
        max_rates=Uniform(100, 300),
        label="Visual",
    )
    channel = nengo.networks.EnsembleArray(
        n_neurons=n_neurons, n_ensembles=dim, label="Channel"
    )
    motor = nengo.networks.EnsembleArray(
        n_neurons=n_neurons, n_ensembles=dim, label="Motor"
    )

    # Creating a memory (integrator)
    tau = 0.1
    memory = nengo.networks.EnsembleArray(
        n_neurons=n_mem, n_ensembles=dim, label="Memory"
    )
    nengo.Connection(memory.output, memory.input, synapse=tau)

    # Function for providing visual input
    def visual_input(t):
        if 0.1 < t < 0.3:
            return vocab.parse("STATEMENT+RED*CIRCLE").v
        elif 0.35 < t < 0.5:
            return vocab.parse("STATEMENT+BLUE*SQUARE").v
        elif 0.55 < t < 0.7:
            return vocab.parse("QUESTION+BLUE").v
        elif 0.75 < t < 0.9:
            return vocab.parse("QUESTION+CIRCLE").v
        return ZERO

    # Function for flipping the output of the thalamus
    def x_biased(x):
        return [1 - x]

    # Providing input to the model
    vis_stim = nengo.Node(output=visual_input, size_out=dim, label="Input stimulus")
    nengo.Connection(vis_stim, visual.input)

    nengo.Connection(visual.output, channel.input, synapse=0.02)
    nengo.Connection(channel.output, memory.input)

    # Creating the unbind network
    unbind = nengo.networks.CircularConvolution(
        n_neurons=n_conv, dimensions=dim, invert_a=True
    )
    nengo.Connection(visual.output, unbind.A)
    nengo.Connection(memory.output, unbind.B)
    nengo.Connection(unbind.output, motor.input)

    # Creating the basal ganglia and the thalamus network
    bg = nengo.networks.BasalGanglia(dimensions=2)
    thal = nengo.networks.Thalamus(dimensions=2)
    nengo.Connection(bg.output, thal.input, synapse=0.01)

    # Defining the transforms for connecting the visual input to the BG
    trans0 = np.matrix(vocab.parse("STATEMENT").v)
    trans1 = np.matrix(vocab.parse("QUESTION").v)
    nengo.Connection(visual.output, bg.input[0], transform=trans0)
    nengo.Connection(visual.output, bg.input[1], transform=trans1)

    # Connecting thalamus output to the two gates gating the channel and the motor
    # populations
    passthrough = nengo.Ensemble(n_neurons, 2)
    nengo.Connection(thal.output, passthrough)

    gate0 = nengo.Ensemble(n_neurons, 1, label="Gate0")
    nengo.Connection(passthrough[0], gate0, function=x_biased, synapse=0.01)
    gate1 = nengo.Ensemble(n_neurons, 1, label="Gate1")
    nengo.Connection(passthrough[1], gate1, function=x_biased, synapse=0.01)

    for ensemble in channel.ea_ensembles:
        nengo.Connection(gate0, ensemble.neurons, transform=[[-3]] * gate0.n_neurons)

    for ensemble in motor.ea_ensembles:
        nengo.Connection(gate1, ensemble.neurons, transform=[[-3]] * gate1.n_neurons)



# flake8: noqa
from .ace_editor import AceEditor
from .component import Component
from .editor import NoEditor
from .htmlview import HTMLView
from .netgraph import NetGraph
from .pointer import Pointer
from .progress import Progress
from .raster import Raster
from .sim_control import SimControl
from .slider import Slider
from .spa_similarity import SpaSimilarity
from .spike_grid import SpikeGrid
from .value import Value
from .voltage import Voltage
from .xyvalue import XYValue

# Old versions of the .cfg files used Templates which had slightly different
# names than the Components currently use.  This code allows us to
# successfully parse those old .cfg files
SliderTemplate = Slider
ValueTemplate = Value
XYValueTemplate = XYValue
SimControlTemplate = SimControl
RasterTemplate = Raster
VoltageTemplate = Voltage
PointerTemplate = Pointer
NetGraphTemplate = NetGraph
AceEditorTemplate = AceEditor



#!/usr/bin/env python
import imp
import io
import os

try:
    from setuptools import find_packages, setup
except ImportError:
    raise ImportError(
        "'setuptools' is required but not installed. To install it, "
        "follow the instructions at "
        "https://pip.pypa.io/en/stable/installing/#installing-with-get-pip-py"
    )


def read(*filenames, **kwargs):
    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


root = os.path.dirname(os.path.realpath(__file__))
version_module = imp.load_source(
    "version", os.path.join(root, "nengo_gui", "version.py")
)

setup(
    name="nengo-gui",
    version=version_module.version,
    author="Applied Brain Research",
    author_email="info@appliedbrainresearch.com",
    packages=find_packages(),
    scripts=[],
    include_package_data=True,
    url="https://github.com/nengo/nengo-gui",
    license="GNU General Public License, version 2",
    description="Web-based GUI for building and visualizing Nengo models.",
    long_description=read("README.rst", "CHANGES.rst"),
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "nengo_gui = nengo_gui:old_main",
            "nengo = nengo_gui:main",
        ]
    },
    install_requires=[
        "nengo>=2.6.0",
    ],
    tests_require=[
        "pytest",
        "selenium",
    ],
    classifiers=[  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Programming Language :: JavaScript",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



# Arbitrary Linear Transformation

# This model shows that any linear transformation between ensembles can be
# easily computed by selecting an appropriate value for the "transform". It
# provides an example of computing linear transformations using vector
# representations.

# Network diagram:  [input - 2D] ---> (x - 2d) ---> (z - 3D)
# A two-dimensional input signal is first fed into a two-dimensional neuronal
# ensemble x, which then projects it on to another three-dimensional ensemble z.

# This model contains the parameters as described in the book. Setting the
# transform equal to the 'weight_matrix' is analogous to entering the weights
# in the "2 to 3 Coupling Matrix" window in Nengo 1.4 GUI as described in the
# book.

# Press the play button to run the simulation.
# The graphs show a two-dimesional input linearly projected on to a
# two-dimensional ensemble of neurons (ens_X), which further linearly projects it on
# to a three-dimesional neuronal ensemble (ens_Z). You can use the sliders to change
# the input values provided by the input node.

import nengo

# Setup the environment
import numpy as np

# Create the network object to which we can add ensembles, connections, etc.
model = nengo.Network(label="Arbitrary Linear Transformation")

with model:
    # Two-dimensional input signal with constant value of [0.5, -0.5]
    stim = nengo.Node([0.5, -0.5], label="Input")

    # 2 and 3-dimensional ensembles each with 200 LIF neurons
    ens_X = nengo.Ensemble(200, dimensions=2, label="X")
    ens_Z = nengo.Ensemble(200, dimensions=3, label="Z")

    # Connect the input to ensemble x
    nengo.Connection(stim, ens_X)

    # Connect ensemble x to ensemble z using a weight matrix
    weight_matrix = [[0.0, 1.0], [1.0, 0.0], [0.5, 0.5]]
    nengo.Connection(ens_X, ens_Z, transform=weight_matrix)



import json
import struct
import time
import timeit

import nengo
import nengo_gui.exec_env
import numpy as np
from nengo_gui.components.component import Component
from nengo_gui.server import WebSocketFrame


class SimControl(Component):
    """Controls simulation via control node embedded in the neural model.

    Also instantiates and communicates with the SimControl and the Toolbar
    on the JavaScript side, which includes the task of back-end selection."""

    config_defaults = dict(shown_time=0.5, kept_time=4.0)

    def __init__(self, dt=0.001):
        # this component must be the very first one defined, so
        # its component_order is the smallest overall
        super(SimControl, self).__init__(component_order=-10)
        self.paused = True
        self.last_tick = None
        self.rate = 0.0
        self.model_dt = dt
        self.rate_tau = 0.5
        self.last_send_rate = None
        self.sim_ticks = 0
        self.skipped = 1
        self.time = 0.0
        self.last_status = None
        self.next_ping_time = None
        self.send_config_options = False
        self.reset_inform = False
        self.node = None
        self.target_rate = 1.0  # desired speed of simulation
        self.target_scale = None  # desired proportion of full speed
        self.delay_time = 0.0  # amount of delay per time step
        self.rate_proportion = 1.0  # current proportion of full speed
        self.smart_sleep_offset = 0.0  # difference from actual sleep time

    def attach(self, page, config, uid):
        super(SimControl, self).attach(page, config, uid)
        self.shown_time = config.shown_time
        self.kept_time = config.kept_time

    def add_nengo_objects(self, page):
        with page.model:
            self.node = nengo.Node(self.control, size_out=0)

    def remove_nengo_objects(self, page):
        page.model.nodes.remove(self.node)

    def finish(self):
        self.page.finish()

    def control(self, t):
        """Node embedded in the model to control simulation progression.

        Sleeps while the simulation is paused.
        """

        self.actual_model_dt = t - self.time
        self.time = t
        self.sim_ticks += 1

        now = timeit.default_timer()
        if self.last_tick is not None:
            dt = now - self.last_tick
            if dt == 0:
                self.skipped += 1
            else:
                rate = self.actual_model_dt * self.skipped / dt
                decay = np.exp(-dt / self.rate_tau)
                self.rate *= decay
                self.rate += (1 - decay) * rate
                self.skipped = 1

                if self.actual_model_dt > 0:
                    # compute current proportion of full speed
                    self.rate_proportion = 1.0 - (
                        (self.rate * self.delay_time) / self.actual_model_dt
                    )

        # if we have a desired proportion, use it to control delay_time
        #  Note that we need last_tick to not be None so that we have a
        #  valid dt value.
        if self.target_scale is not None and self.last_tick is not None:
            s = self.target_scale
            if s <= 0:
                self.delay_time = 0.5
            else:
                self.delay_time = (1.0 / s - s) * (dt - self.delay_time)

        # if we have a desired rate, do a simple P-controller to get there
        if self.target_rate is not None:
            rate_error = self.rate - self.target_rate
            delta = rate_error * 0.0000002
            self.delay_time += delta

        self.delay_time = np.clip(self.delay_time, 0, 0.5)

        if self.delay_time > 0:
            self.smart_sleep(self.delay_time)

        self.last_tick = now

        # Sleeps to prevent the simulation from advancing
        # while the simulation is paused
        while self.paused and self.page.sim is not None:
            time.sleep(0.01)
            self.last_tick = None

    def busy_sleep(self, delay_time):
        now = timeit.default_timer()
        start = now
        while now < start + delay_time:
            now = timeit.default_timer()

    def smart_sleep(self, delay_time):
        """Attempt to sleep for an amount of time without a busy loop.

        This keeps track of the difference between the requested time.sleep()
        time and the actual amount of time slept, and then subtracts that
        difference from future smart_sleep calls.  This should give an
        overall consistent sleep() time even if the actual sleep() time
        is inaccurate.
        """
        t = delay_time + self.smart_sleep_offset
        if t >= 0:
            start = timeit.default_timer()
            time.sleep(t)
            end = timeit.default_timer()
            self.smart_sleep_offset += delay_time - (end - start)
        else:
            self.smart_sleep_offset += delay_time

    def config_settings(self, data):
        for i in data:
            print(i)

    def update_client(self, client):
        now = time.time()
        # send off a ping now and then so we'll notice when connection closes
        if self.next_ping_time is None or now > self.next_ping_time:
            client.write_frame(WebSocketFrame(1, 0, WebSocketFrame.OP_PING, 0, b""))
            self.next_ping_time = now + 2.0

        if self.page.changed:
            self.paused = True
            self.page.sim = None
            self.page.changed = False
        if not self.paused or self.reset_inform:
            client.write_binary(
                struct.pack("<fff", self.time, self.rate, self.rate_proportion)
            )
            self.reset_inform = False
        status = self.get_status()
        if status != self.last_status:
            client.write_text("status:%s" % status)
            self.last_status = status
        if self.send_config_options:
            client.write_text("sims:" + self.backend_options_html())
            client.write_text("config" + "Nengo.Toolbar.prototype.config_modal_show();")
            self.send_config_options = False

    def get_status(self):
        if self.paused:
            return "paused"
        elif self.page.sim is None:
            if self.page.error is None:
                return "building"
            else:
                return "build_error"
        else:
            return "running"

    def javascript(self):
        info = dict(uid=id(self))
        fn = json.dumps(self.page.filename)
        js = self.javascript_config(info)
        return (
            "sim = new Nengo.SimControl(control, %s);\n"
            "toolbar = new Nengo.Toolbar(%s);\n"
            "Nengo.sidemenu = new Nengo.SideMenu();" % (js, fn)
        )

    def message(self, msg):
        if msg == "pause":
            self.paused = True
            if "on_pause" in self.page.locals:
                self.page.locals["on_pause"](self.page.sim)
        elif msg == "config":
            self.send_config_options = True
        elif msg == "continue":
            if self.page.sim is None:
                self.page.rebuild = True
            else:
                if "on_continue" in self.page.locals:
                    self.page.locals["on_continue"](self.page.sim)
            self.paused = False
        elif msg == "reset":
            self.paused = True
            self.time = 0
            self.rate = 0
            self.reset_inform = True
            self.page.sim = None
        elif msg[:8] == "backend:":
            self.page.settings.backend = msg[8:]
            self.page.changed = True
        elif msg[:13] == "target_scale:":
            self.target_scale = float(msg[13:])
            self.target_rate = None

    def backend_options_html(self):
        items = []
        for module in nengo_gui.exec_env.discover_backends():
            if module == self.page.settings.backend:
                selected = " selected"
            else:
                selected = ""
            item = "<option %s>%s</option>" % (selected, module)
            items.append(item)
        return "".join(items)



# Routed Sequencing with Cleanup Memory

# This model extends the routed sequencing model by creating a scalar ensemble
# 'cleanup' as explained in the book. We project the 'state' on to the
# 'cleanup' ensemble using a transformation matrix 'pd', specified by the
# vocabulary vector 'A'. This implies that each component of the state vector
# (i.e., the cortical element) is multiplied by the corresponding component of
# the 'A' vector and summed to produce a one dimensional quantity represented by
# the 'cleanup' ensemble. This operation is similar to the dot product of the
# state vector and the vocabulary vector 'A'.

# This model has parameters as described in the book, with a one dimensional
# cleanup ensemble consisting of 100 neurons.

# Press the play button in the visualizer to run the simulation.
# The graph on the bottom-left shows the semantic pointer representation of the
# values stored in the state and the response of the cleanup population is shown
# in the graph on the bottom-right. The graph on the top-right shows the utility
# (similarity) of the current basal ganglia input (i.e., state) with the
# possible vocabulary vectors. Since the cleanup operation is similar to a dot
# product between the state of the cortical element (state) and the defined
# vocabulary vector 'A', the value of the cleanup population rises only when the
# state (top-right graph) is 'A'.

import nengo

# Setup the environment
import numpy as np
from nengo import spa
from nengo.spa import Vocabulary

# Number of dimensions for the Semantic Pointers
dim = 16

# Change the seed of this RNG to change the vocabulary
rng = np.random.RandomState(4)
vocab = Vocabulary(dimensions=dim, rng=rng, max_similarity=0.1)

# Make a model object with the SPA network
model = spa.SPA(label="Routed_Sequence with cleanupA", vocabs=[vocab])

# Create the spa.SPA network to which we can add SPA objects
with model:
    # Specifying the modules to be used
    model.state = spa.State(dimensions=dim, feedback=1, feedback_synapse=0.01)
    model.vision = spa.State(dimensions=dim)

    # Specify the action mapping
    actions = spa.Actions(
        "dot(vision, START) --> state = vision",
        "dot(state, A) --> state = B",
        "dot(state, B) --> state = C",
        "dot(state, C) --> state = D",
        "dot(state, D) --> state = E",
        "dot(state, E) --> state = A",
    )

    # Creating the BG and thalamus components that confirm to the specified rules
    model.bg = spa.BasalGanglia(actions=actions)
    model.thal = spa.Thalamus(model.bg)

    # Get the transformation matrix (pd) and create the cleanup ensemble (cleanup)
    pd = [model.get_output_vocab("state")["A"].v.tolist()]
    model.cleanup = nengo.Ensemble(n_neurons=100, dimensions=1)

    # Function that provides the model with an initial input semantic pointer.
    def start(t):
        if t < 0.4:
            return "0.8*START+D"
        return "0"

    # Input
    model.input = spa.Input(vision=start)

    # Projecting the state of the cortex on to the cleanup ensemble using a
    # transformation matrix 'pd'.
    nengo.Connection(model.state.output, model.cleanup, transform=pd)
